from __future__ import annotations

from abc import ABC
from collections.abc import Iterator
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
import re

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    ErrorDocument,
    MeasurementType,
    ScanPositionSettingPlateReader,
)
from allotropy.exceptions import (
    AllotropeConversionError,
    get_key_or_error,
)
from allotropy.parsers.constants import NEGATIVE_ZERO
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.pandas import rm_df_columns, SeriesData, set_columns
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    num_to_chars,
    try_float,
    try_float_or_none,
    try_int,
    try_int_or_none,
    try_non_nan_float_or_none,
)

BLOCKS_LINE_REGEX = r"^##BLOCKS=\s*(\d+)$"
END_LINE_REGEX = "~End"
EXPORT_VERSION = "1.3"
VALID_NAN_VALUES = ("Masked", "Range?")


NUM_WELLS_TO_PLATE_DIMENSIONS: dict[int, tuple[int, int]] = {
    6: (3, 2),
    12: (4, 3),
    24: (6, 4),
    48: (8, 6),
    96: (12, 8),
    384: (24, 16),
    1536: (48, 32),
}


def num_wells_to_n_columns(well_count: int) -> int:
    if dimensions := NUM_WELLS_TO_PLATE_DIMENSIONS.get(well_count):
        return dimensions[0]

    num_wells = ",".join(
        [str(num_wells) for num_wells in NUM_WELLS_TO_PLATE_DIMENSIONS]
    )
    msg = (
        f"Unknown number of wells '{well_count}'. Only accepted values are {num_wells}"
    )
    raise AllotropeConversionError(msg)


def time_to_seconds(time: str) -> float:
    """Transforms HH:MM:SS formatted time into seconds"""
    try:
        pt = datetime.strptime(time, "%H:%M:%S").astimezone()
    except ValueError as e:
        msg = "Bad time formatting, expected HH:MM:SS, got {time}"
        raise AllotropeConversionError(msg) from e

    return pt.hour * 3600 + pt.minute * 60 + pt.second


class ReadType(Enum):
    ENDPOINT = "Endpoint"
    KINETIC = "Kinetic"
    SPECTRUM = "Spectrum"
    WELL_SCAN = "Well Scan"


class ExportFormat(Enum):
    TIME_FORMAT = "TimeFormat"
    PLATE_FORMAT = "PlateFormat"


class DataType(Enum):
    RAW = "Raw"
    REDUCED = "Reduced"
    BOTH = "Both"


class ScanPosition(Enum):
    BOTTOM = "Bottom"
    TOP = "Top"
    NONE = None


@dataclass(frozen=True)
class Block:
    block_type: str


@dataclass(frozen=True)
class GroupDataElementEntry:
    name: str
    value: float


@dataclass(frozen=True)
class GroupDataElement:
    sample: str
    positions: list[str]
    plate: str
    entries: list[GroupDataElementEntry]
    errors: list[ErrorDocument]


@dataclass(frozen=True)
class GroupSampleData:
    identifier: str
    data_elements: list[GroupDataElement]

    @classmethod
    def create(cls, data: pd.DataFrame) -> GroupSampleData:
        row_data = [SeriesData(row) for _, row in data.iterrows()]
        top_row = row_data[0]
        identifier = top_row[str, "Sample"]
        data = rm_df_columns(data, r"^Sample$|^Standard Value|^R$|^Unnamed: \d+$")
        # Columns are considered "numeric" if the value of the first row is a float
        # Non-numeric values such as "Mask" and "Range?" will be reported as errors.
        numeric_columns = [
            column
            for column in data.columns
            if top_row.get(float, column, validate=SeriesData.NOT_NAN) is not None
            or top_row.get(str, column) in VALID_NAN_VALUES
        ]

        data_elements: dict[str, list[GroupDataElement]] = {
            column: [] for column in numeric_columns
        }
        for row in row_data:
            row_results = cls._get_entries_and_errors(row, numeric_columns)
            position = row[str, ["Well", "Wells"]]
            plate = row[str, "WellPlateName"]
            for column in numeric_columns:
                if (column_result := row_results.get(column)) is not None:
                    data_elements[column].append(
                        GroupDataElement(
                            sample=identifier,
                            positions=[position],
                            plate=plate,
                            entries=[
                                GroupDataElementEntry(
                                    column,
                                    column_result
                                    if isinstance(column_result, float)
                                    else NEGATIVE_ZERO,
                                )
                            ],
                            errors=[ErrorDocument(column_result, column)]
                            if isinstance(column_result, str)
                            else [],
                        )
                    )
                else:
                    data_elements[column][-1].positions.append(position)

        return GroupSampleData(
            identifier=identifier,
            data_elements=[
                elem for elements in data_elements.values() for elem in elements
            ],
        )

    @classmethod
    def _get_entries_and_errors(
        cls, data_row: SeriesData, column_names: list[str]
    ) -> dict[str, float | str]:
        result: dict[str, float | str] = {}
        for column in column_names:
            value = data_row.get(float, column)
            if value is not None:
                result[column] = value
            elif (error := data_row.get(str, column)) is not None:
                result[column] = error
        return result


@dataclass(frozen=True)
class GroupData:
    name: str
    sample_data: list[GroupSampleData]

    @staticmethod
    def create(reader: CsvReader) -> GroupData:
        name = assert_not_none(
            reader.pop(),
            msg="Unable to find group block name.",
        ).removeprefix("Group: ")

        with pd.option_context("future.no_silent_downcasting", True):  # noqa: FBT003
            data = assert_not_none(
                reader.pop_csv_block_as_df(sep="\t", header=0),
                msg="Unable to find group block data.",
            ).replace(r"^\s+$", None, regex=True)

        assert_not_none(
            data.get("Sample"),
            msg=f"Unable to find sample identifier column in group data {name}",
        )

        samples = data["Sample"].ffill()
        try:
            sample_data = [
                GroupSampleData.create(data.iloc[sample_entries.index])
                for _, sample_entries in samples.groupby(samples)
            ]
        except ValueError as e:
            msg = f"Unable to read Group data format for group {name}."
            raise AllotropeConversionError(msg) from e
        return GroupData(name=name, sample_data=sample_data)


@dataclass(frozen=True)
class GroupColumns:
    data: dict[str, str]

    @staticmethod
    def create(reader: CsvReader) -> GroupColumns:
        data = assert_not_none(
            reader.pop_csv_block_as_df(sep="\t", header=0),
            msg="Unable to find group block columns.",
        )

        if "Formula Name" not in data:
            msg = "Unable to find 'Formula Name' in group block columns."
            raise AllotropeConversionError(msg)

        if "Formula" not in data:
            msg = "Unable to find 'Formula' in group block columns."
            raise AllotropeConversionError(msg)

        return GroupColumns(
            data=dict(zip(data["Formula Name"], data["Formula"], strict=True)),
        )


@dataclass(frozen=True)
class GroupSummaries:
    data: list[str]

    @staticmethod
    def create(reader: CsvReader) -> GroupSummaries:
        data = list(reader.pop_until_empty())
        reader.drop_empty()
        return GroupSummaries(data)


@dataclass(frozen=True)
class GroupBlock(Block):
    group_data: GroupData
    group_columns: GroupColumns
    group_summaries: GroupSummaries

    @staticmethod
    def create(reader: CsvReader) -> GroupBlock:
        return GroupBlock(
            block_type="Group",
            group_data=GroupData.create(reader),
            group_columns=GroupColumns.create(reader),
            group_summaries=GroupSummaries.create(reader),
        )


# TODO do we need to do anything with these?
@dataclass(frozen=True)
class NoteBlock(Block):
    pass


@dataclass(frozen=True)
class PlateHeader:
    name: str
    export_version: str
    export_format: str
    read_type: str
    data_type: str
    kinetic_points: int
    num_wavelengths: int
    wavelengths: list[float]
    num_columns: int
    num_wells: int
    concept: str
    read_mode: str
    unit: str
    read_time: float | None
    read_interval: float | None
    scan_position: ScanPositionSettingPlateReader | None
    reads_per_well: float | None
    pmt_gain: str | None
    num_rows: int
    excitation_wavelengths: list[int] | None
    cutoff_filters: list[int] | None


@dataclass
class DataElement:
    uuid: str
    plate: str
    temperature: float | None
    wavelength: float
    position: str
    value: float
    error_document: list[ErrorDocument]
    elapsed_time: list[float] = field(default_factory=list)
    kinetic_measures: list[float | None] = field(default_factory=list)
    sample_id: str | None = None

    @property
    def sample_identifier(self) -> str:
        return self.sample_id if self.sample_id else f"{self.plate} {self.position}"


@dataclass(frozen=True)
class ReducedDataElement:
    position: str
    value: float


@dataclass(frozen=True)
class PlateWavelengthData:
    wavelength: float
    data_elements: dict[str, DataElement]

    @staticmethod
    def create(
        header: PlateHeader,
        temperature: float | None,
        elapsed_time: float | None,
        wavelength: float,
        df_data: pd.DataFrame,
    ) -> PlateWavelengthData:
        data = {
            f"{num_to_chars(row_idx)}{col}": raw_value
            for row_idx, *row_data in df_data.itertuples()
            for col, raw_value in zip(df_data.columns, row_data, strict=True)
        }

        data_elements = {}
        for position, raw_value in data.items():
            value = try_non_nan_float_or_none(raw_value)
            if value is None and elapsed_time is not None:
                msg = f"Missing kinetic measurement for well position {position} at {elapsed_time}s."
                raise AllotropeConversionError(msg)

            data_elements[str(position)] = DataElement(
                uuid=random_uuid_str(),
                plate=header.name,
                temperature=temperature,
                wavelength=wavelength,
                position=str(position),
                value=NEGATIVE_ZERO if value is None else value,
                error_document=(
                    [ErrorDocument(str(raw_value), header.read_mode)]
                    if value is None
                    else []
                ),
                elapsed_time=[elapsed_time] if elapsed_time is not None else [],
                kinetic_measures=[value] if elapsed_time is not None else [],
            )

        return PlateWavelengthData(wavelength, data_elements)

    def update_kinetic_data_elements(
        self, elapsed_time: float, df_data: pd.DataFrame
    ) -> None:
        data = {
            f"{num_to_chars(row_idx)}{col}": try_non_nan_float_or_none(raw_value)
            for row_idx, *row_data in df_data.itertuples()
            for col, raw_value in zip(df_data.columns, row_data, strict=True)
        }
        for position, value in data.items():
            if value is None:
                msg = f"Missing kinetic measurement for well position {position} at {elapsed_time}s."
                raise AllotropeConversionError(msg)

            self.data_elements[position].elapsed_time.append(elapsed_time)
            self.data_elements[position].kinetic_measures.append(value)


@dataclass(frozen=True)
class RawData:
    wavelength_data: list[PlateWavelengthData]

    @staticmethod
    def get_measurement_section(
        reader: CsvReader, columns: pd.Series[str], rows: int
    ) -> pd.DataFrame:
        lines = []
        # read number of rows in plate section
        for i in range(rows):
            if not (line := reader.pop()):
                msg = f"Expected {rows} rows in measurement table, got {i}."
                raise AllotropeConversionError(msg)
            lines.append(line)
        reader.drop_empty()

        # convert rows to df
        data = assert_not_none(
            reader.lines_as_df(lines=lines, sep="\t"),
            msg="unable to find data from plate block.",
        )
        set_columns(data, columns)
        return data


@dataclass(frozen=True)
class PlateRawData(RawData):
    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> PlateRawData:

        columns = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find data columns for plate block raw data.",
        )

        # use plate dimensions to determine how many rows of plate block to read
        dimensions = assert_not_none(
            NUM_WELLS_TO_PLATE_DIMENSIONS.get(header.num_wells),
            msg="unable to determine plate dimensions",
        )
        rows = dimensions[1]
        data = RawData.get_measurement_section(reader, columns, rows)

        # get temperature and elapsed time (kinetic) from the first column of the first row with value
        first_row_idx = int(pd.to_numeric(data.first_valid_index()))
        temperature = try_non_nan_float_or_none(str(data.iloc[first_row_idx, 1]))
        elapsed_time = None
        if pd.notna(elapsed := data.iloc[first_row_idx, 0]):
            elapsed_time = time_to_seconds(str(elapsed))

        plate_raw_data = PlateRawData(
            wavelength_data=PlateRawData._get_wavelength_data(
                temperature,
                elapsed_time,
                header,
                data.iloc[:, 2:],
            )
        )

        if elapsed_time is not None and header.kinetic_points > 1:
            for _ in range(header.kinetic_points - 1):
                data = RawData.get_measurement_section(reader, columns, rows)

                elapsed_time = time_to_seconds(str(data.iloc[0, 0]))
                plate_raw_data._update_kinetic_data(
                    elapsed_time, header, data.iloc[:, 2:]
                )

        return plate_raw_data

    @staticmethod
    def _get_wavelength_data(
        temperature: float | None,
        elapsed_time: float | None,
        header: PlateHeader,
        w_data: pd.DataFrame,
    ) -> list[PlateWavelengthData]:
        wavelength_data = []
        for idx in range(header.num_wavelengths):
            start = idx * (header.num_columns + 1)
            end = start + header.num_columns
            wavelength_data.append(
                PlateWavelengthData.create(
                    header=header,
                    temperature=temperature,
                    elapsed_time=elapsed_time,
                    wavelength=header.wavelengths[idx],
                    df_data=w_data.iloc[:, start:end],
                )
            )
        return wavelength_data

    def _update_kinetic_data(
        self, elapsed_time: float, header: PlateHeader, w_data: pd.DataFrame
    ) -> None:
        for idx, plate_wavelength_data in enumerate(self.wavelength_data):
            start = idx * (header.num_columns + 1)
            end = start + header.num_columns
            plate_wavelength_data.update_kinetic_data_elements(
                elapsed_time, w_data.iloc[:, start:end]
            )


@dataclass(frozen=True)
class SpectrumRawPlateData(RawData):
    maximum_wavelength_signal: dict[str, float]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> SpectrumRawPlateData:

        columns = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find data columns for plate block raw data.",
        )
        dimensions = assert_not_none(
            NUM_WELLS_TO_PLATE_DIMENSIONS.get(header.num_wells),
            msg="unable to determine plate dimensions",
        )
        rows = dimensions[1]
        wavelength_data: list[PlateWavelengthData] = []

        for wavelength in header.wavelengths:
            data = RawData.get_measurement_section(reader, columns, rows)
            first_row_idx = int(pd.to_numeric(data.first_valid_index()))
            temperature = try_non_nan_float_or_none(str(data.iloc[first_row_idx, 1]))
            wavelength_data.append(
                PlateWavelengthData.create(
                    header=header,
                    temperature=temperature,
                    elapsed_time=None,
                    wavelength=wavelength,
                    df_data=data.iloc[:, 2:],
                )
            )
        reader.pop()
        reader.drop_empty()
        max_wavelength_signal_data = RawData.get_measurement_section(
            reader, columns, rows
        ).iloc[:, 2:]
        signal_data = {
            f"{num_to_chars(row_idx)}{col}": try_float(
                str(raw_value), "wavelength signal"
            )
            for row_idx, *row_data in max_wavelength_signal_data.itertuples()
            for col, raw_value in zip(
                max_wavelength_signal_data.columns, row_data, strict=True
            )
        }

        return SpectrumRawPlateData(
            wavelength_data=wavelength_data,
            maximum_wavelength_signal=signal_data,
        )


@dataclass(frozen=True)
class PlateReducedData:
    data: list[ReducedDataElement]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> PlateReducedData:
        if header.read_type == ReadType.SPECTRUM.value:
            return PlateReducedData(data=[])

        raw_data = assert_not_none(
            reader.pop_csv_block_as_df(sep="\t", header=0),
            msg="Unable to find reduced data for plate block.",
        )

        start = 2
        df_data = raw_data.iloc[:, start : (start + header.num_columns)]

        reduced_data_elements = []
        for row, *data in df_data.itertuples():
            for col, str_value in zip(df_data.columns, data, strict=True):
                value = try_non_nan_float_or_none(str_value)
                if value is not None:
                    reduced_data_elements.append(
                        ReducedDataElement(
                            position=f"{num_to_chars(row)}{col}",
                            value=value,
                        )
                    )
        return PlateReducedData(data=reduced_data_elements)


@dataclass(frozen=True)
class PlateData:
    raw_data: PlateRawData | SpectrumRawPlateData
    reduced_data: PlateReducedData | None

    @staticmethod
    def create(
        reader: CsvReader,
        header: PlateHeader,
    ) -> PlateData:
        raw_data: PlateRawData | SpectrumRawPlateData = (
            SpectrumRawPlateData.create(reader, header)
            if header.read_type == ReadType.SPECTRUM.value
            else PlateRawData.create(reader, header)
        )
        return PlateData(
            raw_data=raw_data,
            reduced_data=(
                PlateReducedData.create(reader, header)
                if reader.current_line_exists()
                else None
            ),
        )

    def iter_data_elements(self, position: str) -> Iterator[DataElement]:
        for wavelength_data in self.raw_data.wavelength_data:
            yield wavelength_data.data_elements[position]


@dataclass(frozen=True)
class TimeMeasurementData:
    data_elements: dict[str, DataElement]

    @staticmethod
    def create(
        header: PlateHeader,
        wavelength: float,
        row: pd.Series[float],
    ) -> TimeMeasurementData:
        temperature = try_non_nan_float_or_none(str(row.iloc[1]))
        data_elements = {}

        for position, raw_value in row.iloc[2:].items():
            value = try_non_nan_float_or_none(raw_value)
            error_document = []
            if value is None:
                error_document.append(ErrorDocument(str(raw_value), header.read_mode))
            data_elements[str(position)] = DataElement(
                uuid=random_uuid_str(),
                plate=header.name,
                temperature=temperature,
                wavelength=wavelength,
                position=str(position),
                value=NEGATIVE_ZERO if value is None else value,
                error_document=error_document,
            )

        return TimeMeasurementData(data_elements)


@dataclass(frozen=True)
class TimeWavelengthData:
    wavelength: float
    measurement_data: list[TimeMeasurementData]

    @staticmethod
    def create(
        reader: CsvReader,
        header: PlateHeader,
        wavelength: float,
        columns: pd.Series[str],
    ) -> TimeWavelengthData:
        data = assert_not_none(
            reader.pop_csv_block_as_df(sep="\t"),
            msg="unable to find raw data from time block.",
        )
        set_columns(data, columns)
        return TimeWavelengthData(
            wavelength=wavelength,
            measurement_data=[
                TimeMeasurementData.create(header, wavelength, row)
                for _, row in data.iterrows()
            ],
        )


@dataclass(frozen=True)
class TimeRawData:
    wavelength_data: list[TimeWavelengthData]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> TimeRawData:
        columns = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find data columns for time block raw data.",
        )

        return TimeRawData(
            wavelength_data=[
                TimeWavelengthData.create(
                    reader,
                    header,
                    wavelength,
                    columns,
                )
                for wavelength in header.wavelengths
            ]
        )


@dataclass(frozen=True)
class TimeReducedData:
    data: list[ReducedDataElement]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> TimeReducedData:
        columns = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find columns for time block reduced data.",
        )
        data = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find reduced data from time block.",
        )
        data.index = pd.Index(columns)

        reduced_data_elements = []
        for pos, str_value in data[2 : header.num_wells + 2].items():
            value = try_non_nan_float_or_none(str_value)
            if value is not None:
                reduced_data_elements.append(
                    ReducedDataElement(
                        position=str(pos),
                        value=value,
                    )
                )
        return TimeReducedData(data=reduced_data_elements)


@dataclass(frozen=True)
class TimeData:
    raw_data: TimeRawData
    reduced_data: TimeReducedData | None

    @staticmethod
    def create(
        reader: CsvReader,
        header: PlateHeader,
    ) -> TimeData:
        return TimeData(
            raw_data=TimeRawData.create(reader, header),
            reduced_data=(
                TimeReducedData.create(reader, header)
                if reader.current_line_exists()
                else None
            ),
        )

    def iter_data_elements(self, position: str) -> Iterator[DataElement]:
        for wavelength_data in self.raw_data.wavelength_data:
            for measurement_data in wavelength_data.measurement_data:
                yield measurement_data.data_elements[position]


@dataclass(frozen=True)
class PlateBlock(ABC, Block):
    header: PlateHeader
    block_data: PlateData | TimeData

    @staticmethod
    def read_header(reader: CsvReader) -> pd.Series[str]:
        raw_header_series = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="Unable to find plate block header.",
        )
        return raw_header_series.astype(str).replace("", None).str.strip()

    @staticmethod
    def get_plate_block_cls(header_series: pd.Series[str]) -> type[PlateBlock]:
        plate_block_cls: dict[str, type[PlateBlock]] = {
            "Absorbance": AbsorbancePlateBlock,
            "Fluorescence": FluorescencePlateBlock,
            "Luminescence": LuminescencePlateBlock,
        }
        read_mode = header_series[5]
        return get_key_or_error("read mode", read_mode, plate_block_cls)

    @property
    def measurement_type(self) -> MeasurementType:
        read_mode_to_measurement_type = {
            "Absorbance": MeasurementType.ULTRAVIOLET_ABSORBANCE,
            "Fluorescence": MeasurementType.FLUORESCENCE,
            "Luminescence": MeasurementType.LUMINESCENCE,
            "Kinetic Absorbance": MeasurementType.ULTRAVIOLET_ABSORBANCE_CUBE_DETECTOR,
            "Kinetic Fluorescence": MeasurementType.FLUORESCENCE_CUBE_DETECTOR,
            "Kinetic Luminescence": MeasurementType.LUMINESCENCE_CUBE_DETECTOR,
        }
        return read_mode_to_measurement_type[self.header.read_mode]

    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        raise NotImplementedError

    @classmethod
    def check_export_version(cls, export_version: str) -> None:
        if export_version != EXPORT_VERSION:
            msg = f"Unsupported export version {export_version}; only {EXPORT_VERSION} is supported."
            raise AllotropeConversionError(msg)

    @classmethod
    def check_read_type(cls, read_type: str) -> None:
        if read_type not in (
            ReadType.ENDPOINT.value,
            ReadType.KINETIC.value,
            ReadType.SPECTRUM.value,
        ):
            msg = f"Only Endpoint, Spectrum or Kinetic measurements can be processed at this time, got: {read_type}"
            raise AllotropeConversionError(msg)

    @classmethod
    def get_read_mode(cls, read_type: str, read_mode_raw: str) -> str:
        return f"{'Kinetic ' if read_type == ReadType.KINETIC.value else ''}{read_mode_raw}"

    @classmethod
    def check_data_type(cls, data_type: str) -> None:
        if data_type not in (DataType.RAW.value, DataType.BOTH.value):
            msg = f"The SoftMax Pro file is required to include either 'Raw' or 'Both' (Raw and Reduced) data for all plates, got {data_type}."
            raise AllotropeConversionError(msg)

    @classmethod
    def check_num_wavelengths(
        cls, wavelengths: list[float], num_wavelengths: int
    ) -> None:
        if len(wavelengths) != num_wavelengths:
            msg = f"Unable to find expected number of wavelength values, expected {num_wavelengths}, found {len(wavelengths)}."
            raise AllotropeConversionError(msg)

    @classmethod
    def get_num_wavelengths(cls, num_wavelengths_raw: str | None) -> int:
        return try_int_or_none(num_wavelengths_raw) or 1

    @classmethod
    def get_wavelengths(cls, wavelengths_str: str | None) -> list[float]:
        if wavelengths_str is None:
            return []
        return [
            try_float(wavelength, "wavelength")
            for wavelength in wavelengths_str.split()
        ]

    @classmethod
    def get_wavelengths_from_start_end_step(
        cls, start: str, end: str, step: str
    ) -> list[float]:
        start_f = try_float(start, "start wavelength")
        end_f = try_float(end, "end wavelength")
        step_f = try_float(step, "wavelength step")

        return [
            start_f + step_f * i for i in range(int((end_f - start_f) / step_f) + 1)
        ]

    def iter_wells(self) -> Iterator[str]:
        cols, rows = NUM_WELLS_TO_PLATE_DIMENSIONS[self.header.num_wells]
        for row in range(rows):
            for col in range(1, cols + 1):
                yield f"{num_to_chars(row)}{col}"

    def iter_data_elements(self, position: str | list[str]) -> Iterator[DataElement]:
        position = [position] if isinstance(position, str) else position
        for p in position:
            yield from self.block_data.iter_data_elements(p)

    def iter_reduced_data(self) -> Iterator[ReducedDataElement]:
        if self.block_data.reduced_data:
            yield from self.block_data.reduced_data.data


@dataclass(frozen=True)
class FluorescencePlateBlock(PlateBlock):
    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        [
            _,  # Plate:
            name,
            export_version,
            export_format,
            read_type,
            read_mode_raw,
            raw_scan_position,
            data_type,
            _,  # Pre-read, always FALSE
            kinetic_points_raw,
            read_time,  # read_time_or_scan_pattern
            read_interval,  # read_interval_or_scan_density
            _,  # start_wavelength
            _,  # end_wavelength
            _,  # wavelength_step
            num_wavelengths_raw,
            wavelengths_str,
            _,  # first_column
            _,  # num_columns
            num_wells_raw,
            excitation_wavelengths_str,
            _,  # cutoff
            cutoff_filters_str,
            _,  # sweep_wave
            _,  # sweep_wavelength
            reads_per_well,
            pmt_gain,
            _,  # start_integration_time
            _,  # end_integration_time
            _,  # first_row
            num_rows,
        ] = header[:31]

        cls.check_export_version(export_version)
        cls.check_read_type(read_type)
        cls.check_data_type(data_type)
        if read_type == ReadType.SPECTRUM.value:
            msg = (
                "Spectrum read type is not currently supported for fluorescence plates."
            )
            raise AllotropeConversionError(msg)

        num_wavelengths = cls.get_num_wavelengths(num_wavelengths_raw)
        wavelengths = cls.get_wavelengths(wavelengths_str)
        cls.check_num_wavelengths(wavelengths, num_wavelengths)

        assert_not_none(
            excitation_wavelengths_str,
            msg="Unable to find excitation wavelengths.",
        )

        excitation_wavelengths = [
            try_int(excitation_wavelength, "excitation wavelengths")
            for excitation_wavelength in excitation_wavelengths_str.split()
        ]

        if len(excitation_wavelengths) != num_wavelengths:
            msg = f"Unable to find expected number of excitation values, expected {num_wavelengths}, found {len(excitation_wavelengths)}"
            raise AllotropeConversionError(msg)

        # cutoff filters is an optional field in the input file
        # if present it contains a list of string numbers separated by spaces
        cutoff_filters = (
            [
                try_int(cutoff_filters, "cutoff filters")
                for cutoff_filters in cutoff_filters_str.split()
            ]
            if cutoff_filters_str is not None
            else None
        )

        # if there are cutoff filters check that the size match number of wavelengths
        if cutoff_filters is not None and len(cutoff_filters) != num_wavelengths:
            msg = f"Unable to find expected number of cutoff filter values, expected {num_wavelengths}, found {len(cutoff_filters)}."
            raise AllotropeConversionError(msg)

        if raw_scan_position == "TRUE":
            scan_position = (
                ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_
            )
        elif raw_scan_position == "FALSE":
            scan_position = (
                ScanPositionSettingPlateReader.top_scan_position__plate_reader_
            )
        else:
            msg = f"{raw_scan_position} is not a valid scan position, expected 'TRUE' or 'FALSE'."
            raise AllotropeConversionError(msg)

        num_wells = try_int(num_wells_raw, "num_wells")
        num_columns = num_wells_to_n_columns(num_wells)

        return PlateHeader(
            name=name,
            export_version=export_version,
            export_format=export_format,
            read_type=read_type,
            data_type=data_type,
            kinetic_points=try_int(kinetic_points_raw, "kinetic_points"),
            num_wavelengths=num_wavelengths,
            wavelengths=wavelengths,
            num_columns=num_columns,
            num_wells=num_wells,
            concept="fluorescence",
            read_mode=cls.get_read_mode(read_type, read_mode_raw),
            unit="RFU",
            read_time=try_float_or_none(read_time),
            read_interval=try_float_or_none(read_interval),
            scan_position=scan_position,
            reads_per_well=try_float(reads_per_well, "reads_per_well"),
            pmt_gain=pmt_gain,
            num_rows=try_int(num_rows, "num_rows"),
            excitation_wavelengths=excitation_wavelengths,
            cutoff_filters=cutoff_filters,
        )


@dataclass(frozen=True)
class LuminescencePlateBlock(PlateBlock):
    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        [
            _,  # Plate:
            name,
            export_version,
            export_format,
            read_type,
            read_mode_raw,
            data_type,
            _,  # Pre-read, always FALSE
            kinetic_points_raw,
            read_time,  # read_time_or_scan_pattern
            read_interval,  # read_interval_or_scan_density
            _,  # start_wavelength
            _,  # end_wavelength
            _,  # wavelength_step
            num_wavelengths_raw,
            wavelengths_str,
            _,  # first_column
            _,  # num_columns,
            num_wells_raw,
            _,  # excitation_wavelengths_str
            _,  # cutoff
            _,  # cutoff_filters_str,
            _,  # sweep_wave
            _,  # sweep_wavelength
            reads_per_well,
            pmt_gain,
            _,  # start_integration_time
            _,  # end_integration_time
            _,  # first_row
            num_rows,
        ] = header[:30]

        cls.check_export_version(export_version)
        cls.check_read_type(read_type)
        cls.check_data_type(data_type)
        if read_type == ReadType.SPECTRUM.value:
            msg = (
                "Spectrum read type is not currently supported for luminescence plates."
            )
            raise AllotropeConversionError(msg)

        num_wavelengths = cls.get_num_wavelengths(num_wavelengths_raw)
        wavelengths = cls.get_wavelengths(wavelengths_str)
        cls.check_num_wavelengths(wavelengths, num_wavelengths)

        num_wells = try_int(num_wells_raw, "num_wells")
        num_columns = num_wells_to_n_columns(num_wells)

        return PlateHeader(
            name=name,
            export_version=export_version,
            export_format=export_format,
            read_type=read_type,
            data_type=data_type,
            kinetic_points=try_int(kinetic_points_raw, "kinetic_points"),
            num_wavelengths=num_wavelengths,
            wavelengths=wavelengths,
            num_columns=num_columns,
            num_wells=num_wells,
            concept="luminescence",
            read_mode=cls.get_read_mode(read_type, read_mode_raw),
            unit="RLU",
            read_time=try_float_or_none(read_time),
            read_interval=try_float_or_none(read_interval),
            scan_position=None,
            reads_per_well=try_int(reads_per_well, "reads_per_well"),
            pmt_gain=pmt_gain,
            num_rows=try_int(num_rows, "num_rows"),
            excitation_wavelengths=None,
            cutoff_filters=None,
        )


@dataclass(frozen=True)
class AbsorbancePlateBlock(PlateBlock):
    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        [
            _,  # Plate:
            name,
            export_version,
            export_format,
            read_type,
            read_mode_raw,
            data_type,
            _,  # Pre-read, always FALSE
            kinetic_points_raw,
            read_time,  # read_time_or_scan_pattern
            read_interval,  # read_interval_or_scan_density
            start_wavelength,  # start_wavelength
            end_wavelength,  # end_wavelength
            wavelength_step,  # wavelength_step
            num_wavelengths_raw,
            wavelengths_str,
            _,  # first_column
            _,  # num_columns
            num_wells_raw,
            _,
            num_rows_raw,
        ] = header[:21]

        cls.check_export_version(export_version)
        cls.check_read_type(read_type)
        cls.check_data_type(data_type)

        num_wavelengths = cls.get_num_wavelengths(num_wavelengths_raw)
        if read_type == ReadType.SPECTRUM.value:
            num_wavelengths = try_int(kinetic_points_raw, "kinetic_points")

        wavelengths = cls.get_wavelengths(wavelengths_str)
        if read_type == ReadType.SPECTRUM.value:
            wavelengths = cls.get_wavelengths_from_start_end_step(
                start_wavelength, end_wavelength, wavelength_step
            )
        cls.check_num_wavelengths(wavelengths, num_wavelengths)

        num_wells = try_int(num_wells_raw, "num_wells")
        num_columns = num_wells_to_n_columns(num_wells)

        return PlateHeader(
            name=name,
            export_version=export_version,
            export_format=export_format,
            read_type=read_type,
            data_type=data_type,
            kinetic_points=try_int(kinetic_points_raw, "kinetic_points"),
            num_wavelengths=num_wavelengths,
            wavelengths=wavelengths,
            num_columns=num_columns,
            num_wells=num_wells,
            concept="absorbance",
            read_mode=cls.get_read_mode(read_type, read_mode_raw),
            unit="mAU",
            read_time=try_float_or_none(read_time),
            read_interval=try_float_or_none(read_interval),
            scan_position=None,
            reads_per_well=None,
            pmt_gain=None,
            num_rows=try_int(num_rows_raw, "num_rows"),
            excitation_wavelengths=None,
            cutoff_filters=None,
        )


@dataclass(frozen=True)
class BlockList:
    plate_blocks: dict[str, PlateBlock]
    group_blocks: list[GroupBlock]

    @staticmethod
    def create(reader: CsvReader) -> BlockList:
        plate_blocks = {}
        group_blocks = []

        for sub_reader in BlockList._iter_blocks_reader(reader):
            if sub_reader.match("^Group"):
                if "WellPlateName" in assert_not_none(
                    sub_reader.get_line(sub_reader.current_line + 1),
                    msg="Unable to get columns from group block",
                ):
                    group_blocks.append(GroupBlock.create(sub_reader))
            elif sub_reader.match("^Plate"):
                header_series = PlateBlock.read_header(sub_reader)
                plate_block_cls = PlateBlock.get_plate_block_cls(header_series)
                header = plate_block_cls.parse_header(header_series)

                export_format_to_data_format = {
                    ExportFormat.TIME_FORMAT.value: TimeData,
                    ExportFormat.PLATE_FORMAT.value: PlateData,
                }
                data_format: type[TimeData] | type[PlateData] = get_key_or_error(
                    "export format", header.export_format, export_format_to_data_format
                )
                block_data = data_format.create(sub_reader, header)

                plate_blocks[header.name] = plate_block_cls(
                    block_type="Plate",
                    header=header,
                    block_data=block_data,
                )
            elif not sub_reader.match("^Note"):
                msg = f"Expected block '{sub_reader.get()}' to start with Group, Plate or Note."
                raise AllotropeConversionError(msg)

        return BlockList(
            plate_blocks=plate_blocks,
            group_blocks=group_blocks,
        )

    @staticmethod
    def _get_n_blocks(reader: CsvReader) -> int:
        start_line = reader.pop() or ""
        if search_result := re.search(BLOCKS_LINE_REGEX, start_line):
            return int(search_result.group(1))
        msg = f"Unrecognized start line, expected a line starting with ##BLOCKS, got {start_line}"
        raise AllotropeConversionError(msg)

    @staticmethod
    def _iter_blocks_reader(reader: CsvReader) -> Iterator[CsvReader]:
        n_blocks = BlockList._get_n_blocks(reader)
        for _ in range(n_blocks):
            yield CsvReader(list(reader.pop_until(END_LINE_REGEX)))
            reader.pop()  # drop end line
            reader.drop_empty()


@dataclass(frozen=True)
class StructureData:
    block_list: BlockList

    @staticmethod
    def create(reader: CsvReader) -> StructureData:
        block_list = BlockList.create(reader)

        # update sample_id if it was reported in the group blocks and include errors from calculated data
        for group_block in block_list.group_blocks:
            for group_sample_data in group_block.group_data.sample_data:
                for group_data_element in group_sample_data.data_elements:
                    plate_block = block_list.plate_blocks[group_data_element.plate]
                    for data_element in plate_block.iter_data_elements(
                        group_data_element.positions
                    ):
                        data_element.sample_id = group_data_element.sample
                        data_element.error_document += group_data_element.errors

        return StructureData(block_list)



""" Constants file for ThermoFisher Qubit 4 Adapter"""

# Instrument Software Details
QUBIT_SOFTWARE = "Qubit 4 software"
MODEL_NUMBER = "Qubit 4"
BRAND_NAME = "Qubit"
PRODUCT_MANUFACTURER = "Thermo Fisher Scientific"
DEVICE_TYPE = "fluorescence detector"

# Errors
VALUE_ERROR = "Unable to find value for column"



from collections import defaultdict

from allotropy.allotrope.schema_mappers.adm.pcr.rec._2024._09.qpcr import SampleRoleType
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import WellItem
from allotropy.parsers.appbio_quantstudio.views import View


class SampleView(View[WellItem]):
    def sort_elements(self, well_items: list[WellItem]) -> dict[str, list[WellItem]]:
        items = defaultdict(list)
        for well_item in well_items:
            items[well_item.sample_identifier].append(well_item)
        return dict(items)


class TargetView(View[WellItem]):
    def sort_elements(self, well_items: list[WellItem]) -> dict[str, list[WellItem]]:
        items = defaultdict(list)
        for well_item in well_items:
            items[well_item.target_dna_description].append(well_item)
        return dict(items)


class TargetRoleView(View[WellItem]):
    def sort_elements(self, well_items: list[WellItem]) -> dict[str, list[WellItem]]:
        items = defaultdict(list)
        for well_item in well_items:
            if well_item.sample_role_type == SampleRoleType.standard_sample_role:
                items[well_item.target_dna_description].append(well_item)
        return dict(items)



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
import io
from pathlib import Path
import re
from typing import Any

from allotropy.allotrope.schema_parser.backup_manager import get_original_path
from allotropy.allotrope.schema_parser.path_util import (
    get_manifest_from_schema_path,
    get_schema_path_from_model_path,
)
from allotropy.allotrope.schema_parser.schema_cleaner import _should_filter_key
from allotropy.allotrope.schema_parser.schema_model import (
    get_all_schema_components,
    get_schema_definitions_mapping,
)
from allotropy.parsers.utils.values import assert_not_none

SCHEMA_DIR_PATH = "src/allotropy/allotrope/schemas"
SHARED_FOLDER_MODULE = "allotropy.allotrope.models.shared"


def _values_equal(value1: Any, value2: Any) -> bool:
    if isinstance(value1, dict):
        return isinstance(value2, dict) and _schemas_equal(value1, value2)
    elif isinstance(value1, list):
        return (
            isinstance(value2, list)
            and len(value1) == len(value2)
            and all(
                _values_equal(v1, v2) for v1, v2 in zip(value1, value2, strict=True)
            )
        )
    else:
        return bool(
            value1 == value2
        )  # typing does not like using == on any to return bool


def _schemas_equal(schema1: dict[str, Any], schema2: dict[str, Any]) -> bool:
    schema1_keys = {key for key in schema1.keys() if not _should_filter_key(key)}
    schema2_keys = {key for key in schema2.keys() if not _should_filter_key(key)}
    return schema1_keys == schema2_keys and all(
        _values_equal(schema1[key], schema2[key]) for key in schema1_keys
    )


def get_shared_schema_info(
    schema: dict[str, Any]
) -> tuple[set[str], dict[str, set[str]]]:
    classes_to_skip = set()
    imports_to_add = defaultdict(set)

    # Get all properties from the schema and all definitions.
    schema_mapping = get_all_schema_components(schema)
    if "$defs" in schema:
        schema_mapping.update(schema["$defs"])
    if "$custom" in schema:
        schema_mapping.update(schema["$custom"])

    # Get schemas defined in models/shared
    shared_schema_mapping = get_schema_definitions_mapping()

    # Match schemas to get classes to remove from the generated code and imports to add.
    for name, component_schema in schema_mapping.items():
        for schema_model in shared_schema_mapping.get(name, []):
            # TODO: log warning visible in script output if schema name is found but no schemas match.
            if _schemas_equal(schema_model.schema, component_schema):
                classes_to_skip.add(schema_model.import_info[1])
                imports_to_add[schema_model.import_info[0]].add(
                    schema_model.import_info[1]
                )

    return classes_to_skip, dict(imports_to_add)


def _to_or_union(types: set[str]) -> str:
    return "|".join(
        sorted(t for t in types if t not in ("", "None"))
        + (["None"] if "None" in types else [])
    )


def _is_or_union(type_string: str, sep: str = "|") -> bool:
    # A type string is an | union if there is a | not nested in brackets.
    nested = 0
    for char in type_string:
        if char == "]":
            nested -= 1
        if char == "[":
            nested += 1
        if char == sep and not nested:
            return True
    return False


def _split_types(type_string: str, sep: str) -> tuple[set[str], int]:
    # Given a string representing some or all of a type specification, returns the types up to the first
    # closing bracket, or all types if no bracket if found. If a closing bracket is found, returns the
    # index of that bracket in the string.
    #
    #    "str|int" -> {"str, "int"}, 6 (6 == length of string)
    #
    #    "str|int]|float" -> {"str, "int"}, 7 (7 == index of ])
    #
    inner_types = [""]
    nested = 0
    for index, char in enumerate(type_string):  # noqa: B007
        if char == "]":
            if nested == 0:
                break
            nested -= 1
        if char == "[":
            nested += 1
        # Build up inner_type, breaking on outermost commas
        if char == sep and not nested:
            inner_types.append("")
        else:
            inner_types[-1] += char
    return set(inner_types), index


def _union_to_or(type_string: str) -> str:
    type_string = type_string.replace("Union", "union")
    while "union" in type_string:
        before, after = type_string.split("union[", 1)
        inner_types, end_index = _split_types(after, ",")
        type_string = before + _to_or_union(inner_types) + after[end_index + 1 :]
        # type_string = before + "|".join(sorted(it for it in inner_types if it)) + after[end_index + 1:]
    return type_string


def _parse_field_types(type_string: str) -> set[str]:
    # Parses a set of types from a dataclass field type specification, e.g.
    #   key: Union[str, int] / int|str -> {int, str}
    #
    # Combined duplicated values recursively. These can happen due to class substitutions, e.g.
    #   item: Union[Type, Type1], where Type1 gets replaced with Type becomes:
    #   item: Union[Type, Type] / Type|Type -> {Type}

    # Convert all unions to | operators, as this makes the rest of the logic easier
    type_string = _union_to_or(type_string)

    # In a union, parse each type recursively to clean up inner types and combine duplicates
    if _is_or_union(type_string):
        return set.union(
            *[_parse_field_types(ts) for ts in _split_types(type_string, "|")[0]]
        )

    # If it is not a union and it doesn't end with a bracket, nothing left to do
    if not type_string.endswith("]"):
        return {type_string}

    identifier, inner = type_string.split("[", 1)
    inner = inner[:-1]

    if identifier.lower() in ("list", "set", "tuple"):
        # Special handling for type specifications with lowercase (typedef) identifiers, which
        # can specify a list of types without a union. e.g.
        #     list[str,int] == list[str|int]
        # Handle this by replacing with an or union
        if _is_or_union(inner, ","):
            inner = _to_or_union(_split_types(inner, ",")[0])
        types = _parse_field_types(inner)
    elif identifier.lower() in ("dict", "mapping"):
        key_type, value_types = inner.split(",", 1)
        return {
            f"{identifier}[{key_type},{_to_or_union(_parse_field_types(value_types))}]"
        }
    elif identifier.lower() == "optional":
        return _parse_field_types(inner) | {"None"}
    else:
        types = _parse_field_types(inner)

    if len(types) > 1:
        types_string = _to_or_union(types)
    else:
        types_string = next(iter(types))
    return {f"{identifier}[{types_string}]"}


@dataclass
class DataclassField:
    """Represents a dataclass field."""

    name: str
    default_value: str | None
    field_types: set[str]

    @staticmethod
    def create(contents: str) -> DataclassField:
        name, content = re.sub(r"\s", "", contents).split(":", maxsplit=1)
        type_string, default_value = (
            content.split("=") if "=" in content else (content, None)
        )
        types = _parse_field_types(assert_not_none(type_string))

        return DataclassField(name, default_value, types)

    @property
    def is_required(self) -> bool:
        return "None" not in self.field_types

    @property
    def contents(self) -> str:
        if len(self.field_types) > 1:
            types = _to_or_union(self.field_types)
        else:
            types = next(iter(self.field_types))
        if self.default_value:
            types = f"{types}={self.default_value}"
        return f"{self.name}: {types}"

    def make_optional(self) -> None:
        self.field_types.add("None")
        self.default_value = "None"

    def merge(self, other: DataclassField) -> DataclassField:
        return DataclassField(
            name=self.name,
            # If default values disagree - this must be a now optional field, make it None
            default_value=self.default_value
            if self.default_value == other.default_value
            else "None",
            # Note that combining required + not required == not required
            # When combining types, drop a general dict[str, Any] type. This happens when the schema
            # does not specify any fields, but that's a mistake.
            field_types={
                type_
                for type_ in self.field_types | other.field_types
                if type_.lower() != "dict[str,any]"
            },
        )


@dataclass
class ClassLines:
    """Represents a set of lines defining a class."""

    lines: list[str]
    class_name: str

    @property
    def base_class_name(self) -> str:
        match = re.match("([A-Za-z]*)[0-9]*$", self.class_name)
        if match:
            return str(match.groups(0)[0])
        return self.class_name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ClassLines):
            return False

        if self.base_class_name != other.base_class_name:
            return False

        contents = "".join(self.lines)
        other_contents = "".join(other.lines)
        other_contents_subbed = re.sub(
            f"{other.class_name}([^0-9])", rf"{self.class_name}\g<1>", other_contents
        )
        return contents == other_contents_subbed

    def __repr__(self) -> str:
        return "".join(self.lines)


@dataclass
class TypeName(ClassLines):
    """Represents a set of lines defining a type name."""

    lines: list[str]
    class_name: str
    types: set[str]

    @staticmethod
    def create(lines: list[str]) -> TypeName:
        match = re.match(
            "(\\S+) = \\(?([^).]+)\\)?", "".join([line.strip() for line in lines])
        )
        if not match:
            msg = f"Could not parse type definition for {''.join(lines)}."
            raise AssertionError(msg)
        class_name = match.groups()[0]
        types = {type_.strip() for type_ in match.groups()[1].split("|")}
        lines = [f"{class_name} = {'|'.join(types)}"]
        return TypeName(lines, class_name, types)

    def should_merge(self, _: DataClassLines | TypeName) -> bool:
        return False

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TypeName):
            return False

        if self.base_class_name != other.base_class_name:
            return False

        if self.types != other.types:
            return False

        return True


@dataclass(eq=False)
class DataClassLines(ClassLines):
    """Represents a set of lines defining a dataclass."""

    parent_class_names: list[str]
    fields: dict[str, DataclassField]
    field_name_order: list[str]

    @staticmethod
    def create(
        name: str,
        parent_class_names: list[str],
        fields: dict[str, DataclassField],
        field_name_order: list[str] | None = None,
        is_frozen: bool | None = False,  # noqa: FBT002
    ) -> DataClassLines:
        # Recreate lines with no whitespace from parsed values
        kwargs = ({"frozen": True} if is_frozen else {}) | {"kw_only": True}
        kwargs_string = ",".join(f"{key}={value}" for key, value in kwargs.items())
        lines = [f"@dataclass({kwargs_string})"]

        class_name_line = f"class {name}"
        if parent_class_names:
            class_name_line += f"({','.join(parent_class_names)})"
        class_name_line += ":"
        lines.append(class_name_line)

        field_name_order = field_name_order or list(fields.keys())
        fixed_field_name_order = [
            field_name
            for field_name in field_name_order
            if fields[field_name].is_required
        ] + [
            field_name
            for field_name in field_name_order
            if not fields[field_name].is_required
        ]

        for field_name in fixed_field_name_order:
            lines.append(f"    {fields[field_name].contents}")
        if not fixed_field_name_order:
            lines.append("    pass")

        return DataClassLines(
            lines=[line + "\n" for line in lines],
            class_name=name,
            parent_class_names=parent_class_names,
            fields=fields,
            field_name_order=fixed_field_name_order,
        )

    @property
    def is_frozen(self) -> bool:
        return "frozen=True" in self.lines[0]

    def has_required_fields(self) -> bool:
        return any(field.is_required for field in self.fields.values())

    def has_optional_fields(self) -> bool:
        return any(not field.is_required for field in self.fields.values())

    def merge_parent(self, parent_class: DataClassLines) -> DataClassLines:
        self.parent_class_names.remove(parent_class.class_name)

        self.fields |= parent_class.fields
        for field_name in parent_class.field_name_order:
            if field_name not in self.field_name_order:
                self.field_name_order.append(field_name)

        return DataClassLines.create(
            self.class_name,
            self.parent_class_names,
            self.fields,
            self.field_name_order,
            self.is_frozen,
        )

    def should_merge(self, other: DataClassLines | TypeName) -> bool:
        if isinstance(other, TypeName):
            return other.types == {self.class_name}

        # Special case where one class is an empty rename of the other.
        if not other.fields and set(other.parent_class_names) == {self.class_name}:
            return True

        # parent classes must match
        if set(self.parent_class_names) != set(other.parent_class_names):
            # Special case for OrderedItem, which sometimes gets combined and sometimes does not.
            if set(self.parent_class_names) | set(other.parent_class_names) != {
                "OrderedItem"
            }:
                return False

        return True

    def merge_similar(self, other: DataClassLines | TypeName) -> DataClassLines:
        # Merge another class that has similar fields with this one.

        if isinstance(other, TypeName):
            if other.types != {self.class_name}:
                msg = f"Can't merge TypeName when not a perfect match: {self.class_name}, {other.class_name}."
                raise AssertionError(msg)
            return self

        # Add fields from the other class to this one.
        for field_name in other.fields:
            if field_name not in self.fields:
                self.fields[field_name] = other.fields[field_name]
                self.fields[field_name].make_optional()
                self.field_name_order.append(field_name)

        # For fields in this class not in the other, mark as optional
        for field_name in self.fields:
            if field_name not in other.fields:
                self.fields[field_name].make_optional()

        # Merge fields by combining types into a single union.
        for field_name in self.fields.keys() & other.fields.keys():
            if self.fields[field_name].contents == other.fields[field_name].contents:
                continue
            self.fields[field_name] = self.fields[field_name].merge(
                other.fields[field_name]
            )

        # Special case for OrderedItem, which sometimes gets combined and sometimes does not.
        # If parents do not match, and it is only ordered item, it is because one class got
        # ordered item merged in, so drop it.
        if set(self.parent_class_names) != set(other.parent_class_names):
            if set(self.parent_class_names) | set(other.parent_class_names) == {
                "OrderedItem"
            }:
                self.parent_class_names = []

        return DataClassLines.create(
            self.class_name,
            self.parent_class_names,
            self.fields,
            self.field_name_order,
            self.is_frozen,
        )


def create_class_lines(lines: list[str]) -> ClassLines:
    is_dataclass = lines[0].startswith("@dataclass")

    # Get the lines that are the class description, including the name and parent classes.
    desc_start = 1 if is_dataclass else 0
    desc_end = desc_start
    # Find the first line that has ":", as the class description may be split over multiple lines if there
    # are multiple/long parent class names.
    while desc_end < len(lines) and ":" not in lines[desc_end]:
        desc_end += 1
    class_description = "".join(
        line.strip("\n") for line in lines[desc_start : desc_end + 1]
    )

    # Get the class name
    match = None
    if class_description.startswith("class"):
        match = re.match("class ([^\\(:]*)", class_description)
    elif " = " in class_description:
        # Match type aliasing, e.g. TClass = str
        match = re.match("(\\S+) =", lines[0])
        return TypeName.create(lines)
    if not match:
        msg = f"Could not determine class name for: {''.join(lines)}."
        raise AssertionError(msg)
    class_name = match.groups()[0]

    if not is_dataclass:
        return ClassLines(lines, class_name)

    is_frozen = "frozen=True" in lines[0]

    # Get parent class names
    parent_class_names = []
    match = re.match(f"class {class_name}\\((.*)\\):", class_description)
    parent_class_names = (
        [name.strip() for name in match.groups()[0].split(",")] if match else []
    )

    # Handle case where dataclass is just a rename of another dataclass. We don't need to do anything with
    # these, so just pass them.
    if "pass" in lines[desc_end + 1] and ":" not in lines[desc_end + 1]:
        return DataClassLines.create(
            name=class_name,
            parent_class_names=parent_class_names,
            fields={},
            field_name_order=[],
            is_frozen=is_frozen,
        )

    # Get fields of the dataclass
    fields: dict[str, DataclassField] = {}
    field_name_order = []
    field_contents = ""
    for line in lines[desc_end + 1 :]:
        if ":" in line and field_contents:
            field = DataclassField.create(field_contents)
            fields[field.name] = field
            field_name_order.append(field.name)
            field_contents = line
        else:
            field_contents += line

    field = DataclassField.create(field_contents)
    fields[field.name] = field
    field_name_order.append(field.name)

    return DataClassLines.create(
        name=class_name,
        parent_class_names=parent_class_names,
        fields=fields,
        field_name_order=field_name_order,
        is_frozen=is_frozen,
    )


class ModelClassEditor:
    """
    Iterates over a file of generated models and rewrites it with modifications, including:

    - Adding manifest to base model
    - Factoring out models found in shared/models
    - Fixing issues with dataclass models that cause issues with python (e.g. dataclass with a required
            field can not inherit from a dataclass with an optional field)
    - Combining some dataclasses to reduce combinatorial explosions. Note that this reduces accuracy at the
            cost of readability. We let the schema enforce correctness in these cases.
    """

    def __init__(
        self,
        manifest: str,
        classes_to_skip: set[str],
        imports_to_add: dict[str, set[str]],
        schema_name: str,
    ):
        self.manifest = manifest
        self.classes_to_skip = classes_to_skip
        self.imports_to_add = imports_to_add
        self.schema_name = schema_name

    def _handle_class_lines(
        self, class_name: str, classes: dict[str, ClassLines]
    ) -> list[str] | None:
        class_lines = classes[class_name]

        # A dataclass with required fields can not inherit from a dataclass with an optional field
        if (
            isinstance(class_lines, DataClassLines)
            and class_lines.has_required_fields()
        ):
            for parent_class_name in class_lines.parent_class_names:
                parent_class = classes[parent_class_name]
                if (
                    isinstance(parent_class, DataClassLines)
                    and parent_class.has_optional_fields()
                ):
                    class_lines = class_lines.merge_parent(parent_class)

        # Add manifest to base Model class.
        if class_lines.class_name == "Model" and "manifest" not in "".join(
            class_lines.lines
        ):
            class_lines.lines.append(f'    manifest: str = "{self.manifest}"\n')
        return class_lines.lines

    def _get_class_lines(
        self, file: io.TextIOBase, existing_lines: list[str] | None = None
    ) -> ClassLines | None:
        # Reads lines for the next class and returns as a ClassLines object.
        lines: list[str] = existing_lines or []
        started = False
        while True:
            line = file.readline()
            # Skip empty lines at the start, return when we hit a newline after the class.
            if line == "\n":
                if started:
                    return create_class_lines(lines)
                else:
                    continue
            # End of file, return the class if we've read one in on this run.
            if not line:
                return create_class_lines(lines) if started else None
            started = True
            lines.append(line)

    def _find_substitutions(
        self, classes: dict[str, ClassLines], class_groups: dict[str, set[str]]
    ) -> tuple[dict[str, str], dict[str, str]]:
        substitutions: dict[str, str] = {}
        renames: dict[str, str] = {}

        for class_group in class_groups.values():
            sorted_class_names = sorted(class_group)
            made_change = False
            for i in range(len(sorted_class_names) - 1):
                class1 = classes[sorted_class_names[i]]
                if not isinstance(class1, DataClassLines | TypeName):
                    continue
                for j in range(i + 1, len(sorted_class_names)):
                    class2 = classes[sorted_class_names[j]]
                    if not isinstance(class2, DataClassLines | TypeName):
                        continue
                    made_change = True
                    # If classes are equal or similar enough to merge, substitute.
                    if class1 == class2:
                        substitutions[class2.class_name] = class1.class_name
                    elif isinstance(class1, DataClassLines) and class1.should_merge(
                        class2
                    ):
                        classes[class1.class_name] = class1.merge_similar(class2)
                        substitutions[class2.class_name] = class1.class_name
                    elif isinstance(class2, DataClassLines) and class2.should_merge(
                        class1
                    ):
                        classes[class2.class_name] = class2.merge_similar(class1)
                        substitutions[class1.class_name] = class2.class_name
                    else:
                        made_change = False

            # If we are done combining classes for this group, rename to set number suffixes
            # as low as possible.
            if not made_change:
                base_class_name = classes[sorted_class_names[0]].base_class_name
                for idx, class_name in enumerate(sorted_class_names):
                    # Don't fix names that are already going to be skipped.
                    if class_name in self.classes_to_skip:
                        continue
                    rename = f"{base_class_name}{idx if idx > 0 else ''}"
                    if rename != class_name:
                        renames[class_name] = rename

        return substitutions, renames

    def modify_file(self, file_contents: str) -> str:
        new_contents: list[str] = []

        f = io.StringIO(file_contents)

        # Scan past comments and import lines.
        while True:
            line = f.readline()
            if line.startswith("#   filename:"):
                new_contents.append(f"#   filename:  {self.schema_name}\n")
            # TODO: this needs work to be more robust.
            elif (
                line == "\n"
                or line.startswith("#")
                or line.startswith("from")
                or line.startswith("import")
            ):
                new_contents.append(line)
            else:
                break

        # We read the first line in the first class, so we need to add it back in
        existing_lines = [line]

        # Remove one newline after default imports
        if self.imports_to_add:
            new_contents = new_contents[:-1]
            # Add new imports
            for module, imports in self.imports_to_add.items():
                new_contents.append(
                    f"from {SHARED_FOLDER_MODULE}.{module} import {', '.join(sorted(imports))}\n"
                )
        else:
            new_contents = new_contents[:-2]

        # Parse classes
        classes_in_order = []
        classes: dict[str, ClassLines] = {}
        class_groups: defaultdict[str, set[str]] = defaultdict(set)
        while True:
            class_lines = self._get_class_lines(f, existing_lines)
            existing_lines = []
            if not class_lines:
                break
            classes_in_order.append(class_lines.class_name)
            classes[class_lines.class_name] = class_lines
            class_groups[class_lines.base_class_name].add(class_lines.class_name)

        # If there are identical/similar classes with numerical suffixes, remove them.
        substitutions, renames = self._find_substitutions(classes, class_groups)

        # Build the new file contents before we do substitutions and look for unused classes.
        for class_name in classes_in_order:
            if class_name in self.classes_to_skip or class_name in substitutions:
                continue
            updated_lines = self._handle_class_lines(class_name, classes)
            if updated_lines:
                new_contents.extend(["\n", "\n", *updated_lines])
        new = "".join(new_contents)

        # Do substitutions
        for class_to_remove in sorted(substitutions.keys(), reverse=True):
            new = re.sub(
                f"{class_to_remove}([^0-9])",
                rf"{substitutions[class_to_remove]}\g<1>",
                new,
            )
        for class_to_remove in sorted(renames.keys(), reverse=True):
            new = re.sub(
                f"{class_to_remove}([^0-9])",
                rf"{renames[class_to_remove]}\g<1>",
                new,
            )

        # Check for classes that are no longer used
        unused_classes: set[str] = set()
        for class_name in classes:
            if class_name == "Model":
                continue
            if not re.search(rf"(?<!class )(?<!\w){class_name}(?!\w)(?! = )", new):
                unused_classes.add(class_name)

        # If we changed anything, run through again to look for iterative removals
        if unused_classes or substitutions or renames:
            self.classes_to_skip = unused_classes
            self.imports_to_add = {}
            return self.modify_file(new)

        return new


def modify_file(model_path: Path, schema: dict[str, Any]) -> None:
    classes_to_skip, imports_to_add = get_shared_schema_info(schema)
    schema_path = get_schema_path_from_model_path(get_original_path(model_path))
    manifest = get_manifest_from_schema_path(schema_path)
    editor = ModelClassEditor(
        manifest, classes_to_skip, imports_to_add, schema_path.name
    )

    with open(model_path) as f:
        contents = f.read()

    with open(model_path, "w") as f:
        f.write(editor.modify_file(contents))



from __future__ import annotations

from pathlib import Path

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    CalculatedDataItem,
    DataSource,
    ErrorDocument,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedDataDocument,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import NEGATIVE_ZERO, NOT_APPLICABLE
from allotropy.parsers.unchained_labs_lunatic.constants import (
    CALCULATED_DATA_LOOKUP,
    DETECTION_TYPE,
    DEVICE_TYPE,
    INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG,
    MODEL_NUMBER,
    NO_DATE_OR_TIME_ERROR_MSG,
    NO_DEVICE_IDENTIFIER_ERROR_MSG,
    NO_MEASUREMENT_IN_PLATE_ERROR_MSG,
    NO_WAVELENGTH_COLUMN_ERROR_MSG,
    PRODUCT_MANUFACTURER,
    SOFTWARE_NAME,
    WAVELENGTH_COLUMNS_RE,
)
from allotropy.parsers.utils.pandas import (
    map_rows,
    SeriesData,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
)


def _create_measurement(
    well_plate_data: SeriesData,
    header: SeriesData,
    wavelength_column: str,
    calculated_data: list[CalculatedDataItem],
) -> Measurement:
    if wavelength_column not in well_plate_data.series:
        msg = NO_MEASUREMENT_IN_PLATE_ERROR_MSG.format(wavelength_column)
        raise AllotropeConversionError(msg)

    wavelength_match = WAVELENGTH_COLUMNS_RE.match(wavelength_column)
    if not wavelength_match:
        raise AllotropeConversionError(INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG)
    if len(wavelength_match.groups()) > 1:
        wavelength, path_length = wavelength_match.groups()
    else:
        wavelength = wavelength_match.groups()[0]
        path_length = None

    background_wavelength = well_plate_data.get(float, "Background Wvl. (nm)")
    background_absorbance = None
    if background_wavelength is not None:
        background_absorbance = well_plate_data.get(
            float, f"Background (A{int(background_wavelength)})"
        )

    measurement_identifier = random_uuid_str()
    error_documents: list[ErrorDocument] = []
    calculated_data.extend(
        _get_calculated_data(
            well_plate_data, wavelength_column, measurement_identifier, error_documents
        )
    )
    absorbance = well_plate_data.get(float, wavelength_column)

    if absorbance is None:
        error_documents.append(
            ErrorDocument(
                error=NOT_APPLICABLE,
                error_feature=DETECTION_TYPE.lower(),
            )
        )

    concentration_factor = well_plate_data.get(float, "Concentration factor (ng/ul)")
    return Measurement(
        type_=MeasurementType.ULTRAVIOLET_ABSORBANCE,
        device_type=DEVICE_TYPE,
        detection_type=DETECTION_TYPE,
        identifier=measurement_identifier,
        detector_wavelength_setting=float(wavelength),
        electronic_absorbance_reference_wavelength_setting=background_wavelength,
        measurement_custom_info={
            "electronic_absorbance_reference_absorbance": background_absorbance
        },
        absorbance=absorbance if absorbance is not None else NEGATIVE_ZERO,
        sample_identifier=well_plate_data[str, "Sample name"],
        location_identifier=well_plate_data[str, "Plate Position"],
        well_plate_identifier=well_plate_data.get(str, "Plate ID"),
        batch_identifier=well_plate_data.get(str, "Sample Group"),
        firmware_version=header.get(str, "Client version"),
        sample_custom_info={
            "path length": float(path_length) if path_length is not None else None,
        },
        device_control_custom_info={
            "path length mode": well_plate_data.get(str, "Path length mode"),
            "pump": well_plate_data.get(str, "Pump"),
        },
        error_document=error_documents,
        processed_data_document=ProcessedDataDocument(
            identifier=random_uuid_str(), concentration_factor=concentration_factor
        )
        if concentration_factor is not None
        else None,
    )


def _get_calculated_data(
    well_plate_data: SeriesData,
    wavelength_column: str,
    measurement_identifier: str,
    error_documents: list[ErrorDocument],
) -> list[CalculatedDataItem]:
    calculated_data = []
    for item in CALCULATED_DATA_LOOKUP.get(wavelength_column, []):
        value = well_plate_data.get(float, item["column"])
        if value is None:
            error_documents.append(
                ErrorDocument(
                    error=NOT_APPLICABLE,
                    error_feature=item["name"],
                )
            )
            continue

        calculated_data.append(
            CalculatedDataItem(
                identifier=random_uuid_str(),
                name=item["name"],
                value=value,
                unit=item["unit"],
                data_sources=[
                    DataSource(
                        identifier=measurement_identifier,
                        feature=item["feature"],
                    )
                ],
            )
        )
    return calculated_data


def _create_measurement_group(
    data: SeriesData,
    wavelength_columns: list[str],
    calculated_data: list[CalculatedDataItem],
    header: SeriesData,
) -> MeasurementGroup:
    timestamp = header.get(str, "Date")
    # Support timestamp from metadata section, but overide with columns in data if specified.
    date = data.get(str, "Date")
    time = data.get(str, "Time")
    if date and time:
        timestamp = f"{date} {time}"

    return MeasurementGroup(
        measurement_time=assert_not_none(timestamp, msg=NO_DATE_OR_TIME_ERROR_MSG),
        analyst=header.get(str, "Test performed by"),
        analytical_method_identifier=data.get(str, "Application"),
        experimental_data_identifier=header.get(str, "Experiment name"),
        plate_well_count=96,
        measurements=[
            _create_measurement(data, header, wavelength_column, calculated_data)
            for wavelength_column in wavelength_columns
        ],
    )


def create_metadata(header: SeriesData, file_path: str) -> Metadata:
    asm_file_identifier = Path(file_path).with_suffix(".json")
    device_identifier = header.get(str, "Instrument ID") or header.get(
        str, "Instrument"
    )
    return Metadata(
        model_number=MODEL_NUMBER,
        product_manufacturer=PRODUCT_MANUFACTURER,
        device_identifier=assert_not_none(
            device_identifier, msg=NO_DEVICE_IDENTIFIER_ERROR_MSG
        ),
        software_name=SOFTWARE_NAME,
        software_version=header.get(str, "Software version"),
        file_name=Path(file_path).name,
        unc_path=file_path,
        asm_file_identifier=asm_file_identifier.name,
        data_system_instance_id=NOT_APPLICABLE,
    )


def create_measurement_groups(
    header: SeriesData, data: pd.DataFrame
) -> tuple[list[MeasurementGroup], list[CalculatedDataItem]]:
    wavelength_columns = list(filter(WAVELENGTH_COLUMNS_RE.match, data.columns))
    if not wavelength_columns:
        raise AllotropeConversionError(NO_WAVELENGTH_COLUMN_ERROR_MSG)

    # TODO: we are reporting calculated data for measurements globally instead of in the measurement doc,
    # which is why we have to pass this list to collect them. Why are we reporting globally when data is
    # pertains to the individual measurements?
    calculated_data: list[CalculatedDataItem] = []

    def make_group(data: SeriesData) -> MeasurementGroup:
        return _create_measurement_group(
            data, wavelength_columns, calculated_data, header
        )

    return map_rows(data, make_group), calculated_data



from abc import ABC, abstractmethod
import uuid


class IdGenerator(ABC):
    @abstractmethod
    def generate_id(self) -> str:
        raise NotImplementedError


class _RandomUuidGenerator(IdGenerator):
    def generate_id(self) -> str:
        return str(uuid.uuid4())


_ID_GENERATOR = _RandomUuidGenerator()


class _IdGeneratorFactory:
    def get_id_generator(self) -> IdGenerator:
        return _ID_GENERATOR


__ID_GENERATOR_FACTORY = _IdGeneratorFactory()


def random_uuid_str() -> str:
    return __ID_GENERATOR_FACTORY.get_id_generator().generate_id()



from dataclasses import dataclass
from enum import Enum
from typing import Any

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.binding_affinity_analyzer.wd._2024._12.binding_affinity_analyzer import (
    BindingAffinityAnalyzerAggregateDocument,
    BindingAffinityAnalyzerDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocument,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    SensorChipDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueMicroliterPerMinute,
    TQuantityValueMolar,
    TQuantityValueNanomolar,
    TQuantityValuePercent,
    TQuantityValuePerSecond,
    TQuantityValueSecondTime,
    TQuantityValueTODO,
)
from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.utils.values import assert_not_none, quantity_or_none


class MeasurementType(Enum):
    SURFACE_PLASMON_RESONANCE = "SURFACE_PLASMON_RESONANCE"


@dataclass(frozen=True)
class DataSource:
    identifier: str
    feature: str


@dataclass(frozen=True)
class CalculatedDataItem:
    identifier: str
    name: str
    value: float
    unit: str
    data_sources: list[DataSource]
    description: str | None = None


@dataclass(frozen=True)
class DeviceDocument:
    device_type: str
    device_identifier: str


@dataclass(frozen=True)
class Metadata:
    device_identifier: str
    asm_file_identifier: str
    data_system_instance_identifier: str
    model_number: str
    sensor_chip_identifier: str
    brand_name: str | None = None
    product_manufacturer: str | None = None
    software_name: str | None = None
    software_version: str | None = None
    file_name: str | None = None
    unc_path: str | None = None
    device_document: list[DeviceDocument] | None = None
    detection_type: str | None = None
    equipment_serial_number: str | None = None
    compartment_temperature: float | None = None
    sensor_chip_type: str | None = None
    lot_number: str | None = None
    sensor_chip_custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class Measurements:
    identifier: str
    sample_identifier: str
    device_type: str
    type_: MeasurementType
    sensorgram_posix_path: str
    location_identifier: str | None = None
    batch_identifier: str | None = None
    well_plate_identifier: str | None = None
    sample_role_type: str | None = None
    concentration: JsonFloat | None = None
    method_name: str | None = None
    ligand_identifier: str | None = None
    flow_cell_identifier: str | None = None
    flow_path: str | None = None
    flow_rate: float | None = None
    contact_time: float | None = None
    dilution: float | None = None
    device_control_custom_info: dict[str, Any] | None = None
    binding_on_rate_measurement: JsonFloat | None = None
    binding_off_rate_measurement: JsonFloat | None = None
    equilibrium_dissociation_constant: JsonFloat | None = None
    Rmax: JsonFloat | None = None
    sensorgram_posix_path_identifier: str | None = None
    rpoint_posix_path: str | None = None
    rpoint_posix_path_identifier: str | None = None
    sample_custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurement_time: str
    measurements: list[Measurements]
    experiment_type: str | None = None
    analytical_method_identifier: str | None = None
    analyst: str | None = None
    measurement_aggregate_custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]
    calculated_data: list[CalculatedDataItem] | None = None


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST: str = "http://purl.allotrope.org/manifests/binding-affinity-analyzer/WD/2024/12/binding-affinity-analyzer.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            binding_affinity_analyzer_aggregate_document=BindingAffinityAnalyzerAggregateDocument(
                data_system_document=DataSystemDocument(
                    ASM_file_identifier=data.metadata.asm_file_identifier,
                    data_system_instance_identifier=data.metadata.data_system_instance_identifier,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    software_version=data.metadata.software_version,
                    software_name=data.metadata.software_name,
                ),
                device_system_document=DeviceSystemDocument(
                    device_identifier=data.metadata.device_identifier,
                    model_number=data.metadata.model_number,
                    brand_name=data.metadata.brand_name,
                    product_manufacturer=data.metadata.product_manufacturer,
                    device_document=(
                        [
                            DeviceDocumentItem(
                                device_type=device_document_item.device_type,
                                device_identifier=device_document_item.device_identifier,
                            )
                            for device_document_item in data.metadata.device_document
                        ]
                        if data.metadata.device_document is not None
                        else None
                    ),
                ),
                binding_affinity_analyzer_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
            )
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> BindingAffinityAnalyzerDocumentItem:
        return BindingAffinityAnalyzerDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=add_custom_information_document(
                MeasurementAggregateDocument(
                    measurement_time=self.get_date_time(
                        assert_not_none(measurement_group.measurement_time)
                    ),
                    experiment_type=measurement_group.experiment_type,
                    analytical_method_identifier=measurement_group.analytical_method_identifier,
                    compartment_temperature=quantity_or_none(
                        TQuantityValueDegreeCelsius, metadata.compartment_temperature
                    ),
                    measurement_document=[
                        self._get_measurement_document_item(measurements, metadata)
                        for measurements in measurement_group.measurements
                    ],
                ),
                custom_info_doc=measurement_group.measurement_aggregate_custom_info,
            ),
        )

    def _get_measurement_document_item(
        self, measurement: Measurements, metadata: Metadata
    ) -> MeasurementDocument:
        if measurement.type_ == MeasurementType.SURFACE_PLASMON_RESONANCE:
            return self._get_surface_plasmon_resonance_measurement_document(
                measurement, metadata
            )
        else:
            msg = f"Invalid measurement type: {measurement.type_}"
            raise AllotropyParserError(msg)

    def _get_surface_plasmon_resonance_measurement_document(
        self, measurements: Measurements, metadata: Metadata
    ) -> MeasurementDocument:
        return MeasurementDocument(
            measurement_identifier=measurements.identifier,
            POSIX_path=measurements.sensorgram_posix_path,
            identifier=measurements.sensorgram_posix_path_identifier,
            sample_document=add_custom_information_document(
                SampleDocument(
                    sample_identifier=measurements.sample_identifier,
                    sample_role_type=measurements.sample_role_type,
                    location_identifier=measurements.location_identifier,
                    concentration=quantity_or_none(
                        TQuantityValueNanomolar, measurements.concentration
                    ),
                ),
                custom_info_doc=measurements.sample_custom_info,
            ),
            detection_type=metadata.detection_type,
            method_name=measurements.method_name,
            ligand_identifier=measurements.ligand_identifier,
            sensor_chip_document=add_custom_information_document(
                SensorChipDocument(
                    sensor_chip_identifier=metadata.sensor_chip_identifier,
                    sensor_chip_type=metadata.sensor_chip_type,
                    lot_number=metadata.lot_number,
                    product_manufacturer=metadata.product_manufacturer,
                ),
                custom_info_doc=metadata.sensor_chip_custom_info,
            ),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    add_custom_information_document(
                        DeviceControlDocumentItem(
                            flow_cell_identifier=measurements.flow_cell_identifier,
                            flow_path=measurements.flow_path,
                            flow_rate=quantity_or_none(
                                TQuantityValueMicroliterPerMinute,
                                measurements.flow_rate,
                            ),
                            contact_time=quantity_or_none(
                                TQuantityValueSecondTime, measurements.contact_time
                            ),
                            dilution_factor=quantity_or_none(
                                TQuantityValuePercent, measurements.dilution
                            ),
                            device_type=measurements.device_type,
                        ),
                        custom_info_doc=measurements.device_control_custom_info,
                    )
                ]
            ),
            processed_data_aggregate_document=ProcessedDataAggregateDocument(
                processed_data_document=[
                    ProcessedDataDocumentItem(
                        POSIX_path=measurements.rpoint_posix_path,
                        identifier=measurements.rpoint_posix_path_identifier,
                        binding_on_rate_measurement_datum__kon_=quantity_or_none(
                            TQuantityValueTODO, measurements.binding_on_rate_measurement
                        ),
                        binding_off_rate_measurement_datum__koff_=quantity_or_none(
                            TQuantityValuePerSecond,
                            measurements.binding_off_rate_measurement,
                        ),
                        equilibrium_dissociation_constant__KD_=quantity_or_none(
                            TQuantityValueMolar,
                            measurements.equilibrium_dissociation_constant,
                        ),
                        maximum_binding_capacity__Rmax_=quantity_or_none(
                            TQuantityValueTODO,
                            measurements.Rmax,
                        ),
                    )
                ]
            ),
        )



from allotropy.allotrope.models.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.cytiva_unicorn.constants import DISPLAY_NAME
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.cytiva_unicorn.structure.measurement_group import (
    create_measurement_groups,
)
from allotropy.parsers.cytiva_unicorn.structure.metadata import (
    create_metadata,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class CytivaUnicornParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "zip"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        handler = UnicornZipHandler(named_file_contents.get_bytes_stream())
        results = handler.get_results()
        return Data(
            create_metadata(handler, results, named_file_contents.original_file_path),
            create_measurement_groups(handler, results),
        )



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import re
from typing import Any

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.solution_analyzer.rec._2024._09.solution_analyzer import (
    Analyte,
    DataProcessing,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.novabio_flex2.constants import (
    ANALYTE_MAPPINGS,
    DATA_PROCESSING_FIELDS,
    DETECTION_PROPERTY_MAPPING,
    DEVICE_TYPE,
    FILENAME_REGEX,
    INVALID_FILENAME_MESSAGE,
    MODEL_NUMBER,
    PRODUCT_MANUFACTURER,
    SOFTWARE_NAME,
)
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import try_float_or_none


@dataclass(frozen=True)
class Title:
    processing_time: str
    device_identifier: str | None

    @staticmethod
    def create(file_path: str) -> Title:
        filename = Path(file_path).name
        matches = re.match(FILENAME_REGEX, filename, flags=re.IGNORECASE)

        if not matches:
            raise AllotropeConversionError(INVALID_FILENAME_MESSAGE.format(filename))

        matches_dict = matches.groupdict()
        return Title(
            processing_time=matches_dict["processing_time"].replace("_", " "),
            device_identifier=matches_dict["device_identifier"] or None,
        )


@dataclass(frozen=True)
class Sample:
    identifier: str
    sample_type: str
    measurement_time: str
    batch_identifier: str | None
    analytes: list[Analyte]
    po2: float | None = None
    pco2: float | None = None
    carbon_dioxide_saturation: float | None = None
    oxygen_saturation: float | None = None
    ph: float | None = None
    temperature: float | None = None
    osmolality: float | None = None
    viability: float | None = None
    total_cell_density: float | None = None
    viable_cell_density: float | None = None
    average_live_cell_diameter: float | None = None
    total_cell_count: float | None = None
    viable_cell_count: float | None = None
    cell_type_processing_method: str | None = None
    cell_density_dilution_factor: float | None = None

    @classmethod
    def create(cls, series: pd.Series[Any]) -> Sample:
        data = SeriesData(series)
        cell_density_dilution = data.get(str, "Cell Density Dilution", "")
        if cell_density_dilution:
            cell_density_dilution = cell_density_dilution.split(":")[0]

        return Sample(
            identifier=data[str, "Sample ID"],
            sample_type=data[str, "Sample Type"],
            measurement_time=data[str, "Date & Time"],
            batch_identifier=data.get(str, "Batch ID"),
            analytes=sorted(
                [
                    Analyte(
                        ANALYTE_MAPPINGS[raw_name]["name"],
                        data[float, raw_name],
                        ANALYTE_MAPPINGS[raw_name]["unit"],
                    )
                    for raw_name in ANALYTE_MAPPINGS
                    if data.get(float, raw_name) is not None
                ]
            ),
            viability=data.get(float, "Viability"),
            total_cell_density=data.get(float, "Total Density"),
            viable_cell_density=data.get(float, "Viable Density"),
            average_live_cell_diameter=data.get(float, "Average Live Cell Diameter"),
            total_cell_count=data.get(float, "Total Cell Count"),
            viable_cell_count=data.get(float, "Total Live Count"),
            osmolality=data.get(float, "Osm"),
            ph=data.get(float, "pH"),
            temperature=data.get(float, "Vessel Temperature (°C)"),
            po2=data.get(float, "PO2"),
            pco2=data.get(float, "PCO2"),
            carbon_dioxide_saturation=data.get(float, "CO2 Saturation"),
            oxygen_saturation=data.get(float, "O2 Saturation"),
            cell_type_processing_method=data.get(str, "Cell Type")
            if cell_density_dilution
            else None,
            cell_density_dilution_factor=try_float_or_none(str(cell_density_dilution)),
        )


@dataclass(frozen=True)
class SampleList:
    analyst: str
    samples: list[Sample]

    @staticmethod
    def create(data: pd.DataFrame) -> SampleList:
        sample_data_rows = [row for _, row in data.iterrows()]

        if not sample_data_rows:
            msg = "Unable to find any sample."
            raise AllotropeConversionError(msg)

        analyst = sample_data_rows[0].get("Operator")

        if analyst is None:
            msg = "Unable to find the Operator."
            raise AllotropeConversionError(msg)

        return SampleList(
            analyst=str(analyst),
            samples=[Sample.create(sample_data) for sample_data in sample_data_rows],
        )


def _create_measurement(sample: Sample, **kwargs: Any) -> Measurement:
    return Measurement(
        identifier=random_uuid_str(),
        measurement_time=sample.measurement_time,
        sample_identifier=sample.identifier,
        description=sample.sample_type,
        batch_identifier=sample.batch_identifier,
        **kwargs,
    )


def _get_measurements(sample: Sample) -> list[Measurement]:
    measurements = []

    # NOTE: only specifying this order to keep test results identical for refactor. Will remove in follow
    # up change that moves this logic into schema mapper.
    for detection_type in [
        "metabolite-detection",
        "cell-counting",
        "blood-gas-detection",
        "osmolality-detection",
        "ph-detection",
    ]:
        kwargs = {
            key: getattr(sample, key)
            for key in DETECTION_PROPERTY_MAPPING[detection_type]
        }
        data_processing = {
            key: value
            for key in DATA_PROCESSING_FIELDS
            if key in kwargs and (value := kwargs.pop(key)) is not None
        }
        if data_processing:
            kwargs["data_processing"] = DataProcessing(**data_processing)
        if any(value is not None for value in kwargs.values()):
            measurements.append(
                _create_measurement(
                    sample,
                    detection_type=detection_type,
                    **kwargs,
                )
            )

    return measurements


def create_metadata(title: Title, file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        file_name=path.name,
        unc_path=file_path,
        device_type=DEVICE_TYPE,
        model_number=MODEL_NUMBER,
        product_manufacturer=PRODUCT_MANUFACTURER,
        device_identifier=title.device_identifier,
        software_name=SOFTWARE_NAME,
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_identifier=NOT_APPLICABLE,
    )


def create_measurement_groups(
    title: Title, sample_list: SampleList
) -> list[MeasurementGroup]:
    return [
        MeasurementGroup(
            analyst=sample_list.analyst,
            data_processing_time=title.processing_time,
            measurements=_get_measurements(sample),
        )
        for sample in sample_list.samples
    ]



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
import re
from typing import Any, TypeVar

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.pcr.rec._2024._09.qpcr import SampleRoleType
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_reader import (
    AppBioQuantStudioReader,
)
from allotropy.parsers.appbio_quantstudio.constants import (
    ExperimentType,
    SAMPLE_ROLE_TYPES_MAP,
)
from allotropy.parsers.constants import NEGATIVE_ZERO, NOT_APPLICABLE
from allotropy.parsers.utils.calculated_data_documents.definition import Referenceable
from allotropy.parsers.utils.pandas import df_to_series_data, map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none, try_int

T = TypeVar("T")


def map_wells(map_func: Callable[..., T], data: pd.DataFrame) -> dict[int, T]:
    return {
        try_int(str(well_id), "well id"): map_func(well_data)
        for well_id, well_data in data.groupby("Well")
    }


def get_well_volume(block_type: str) -> float:
    # if the block type includes the well volume in mL, convert to microliters
    if well_search := re.search(r"([0-9]+\.[0-9]+)-?mL", block_type):
        return float(well_search.groups()[0]) * 1000
    # The 384-well block and Taqman Array Card have specified well volumes from the
    # manufacturer (Thermo Fisher) which are 40uL and 1.5uL, respectively.
    elif "384-Well Block" in block_type:
        return 40
    elif "Taqman Array Card" in block_type:
        return 1.5
    # Since well volume is required, if it cannot be implied from block type
    # a negative zero will be returned, indicating an error.
    return NEGATIVE_ZERO


@dataclass(frozen=True)
class Header:
    measurement_time: str
    plate_well_count: int | None
    experiment_type: ExperimentType
    device_identifier: str
    model_number: str
    device_serial_number: str
    measurement_method_identifier: str
    pcr_detection_chemistry: str | None
    passive_reference_dye_setting: str | None
    barcode: str | None
    analyst: str | None
    experimental_data_identifier: str
    well_volume: float

    @staticmethod
    def create(data: SeriesData) -> Header:
        experiments_type_options = {
            "Standard Curve": ExperimentType.standard_curve_qpcr_experiment,
            "Relative Standard Curve": ExperimentType.relative_standard_curve_qpcr_experiment,
            "Comparative Cт (ΔΔCт)": ExperimentType.comparative_ct_qpcr_experiment,
            "Melt Curve": ExperimentType.melt_curve_qpcr_experiment,
            "Genotyping": ExperimentType.genotyping_qpcr_experiment,
            "Presence/Absence": ExperimentType.presence_absence_qpcr_experiment,
        }

        plate_well_count = None
        block_type = data.get(str, "Block Type")
        if block_type:
            plate_well_count_search = re.search("(96)|(384)", block_type)
            if plate_well_count_search:
                plate_well_count = int(plate_well_count_search.group())

        return Header(
            measurement_time=data[str, "Experiment Run End Time"],
            plate_well_count=plate_well_count,
            experiment_type=assert_not_none(
                experiments_type_options.get(
                    data[str, "Experiment Type"].strip(),
                ),
                msg="Unable to find valid experiment type",
            ),
            device_identifier=(data.get(str, "Instrument Name", NOT_APPLICABLE)),
            model_number=data[str, "Instrument Type"],
            device_serial_number=data.get(
                str, "Instrument Serial Number", NOT_APPLICABLE
            ),
            measurement_method_identifier=data[str, "Quantification Cycle Method"],
            pcr_detection_chemistry=data.get(str, "Chemistry"),
            passive_reference_dye_setting=data.get(str, "Passive Reference"),
            barcode=data.get(str, "Experiment Barcode"),
            analyst=data.get(str, "Experiment User Name"),
            experimental_data_identifier=data[str, "Experiment Name"],
            well_volume=get_well_volume(block_type) if block_type else NEGATIVE_ZERO,
        )


@dataclass
class WellItem(Referenceable):
    identifier: int
    target_dna_description: str
    sample_identifier: str
    location_identifier: str
    reporter_dye_setting: str | None = None
    position: str | None = None
    well_location_identifier: str | None = None
    quencher_dye_setting: str | None = None
    sample_role_type: SampleRoleType | None = None
    group_identifier: str | None = None
    extra_data: dict[str, Any] | None = None
    _result: Result | None = None

    # Make hashable to allow for use of caching
    def __hash__(self) -> int:
        return hash(self.identifier)

    @property
    def has_result(self) -> bool:
        return self._result is not None

    @property
    def result(self) -> Result:
        return assert_not_none(self._result)

    @staticmethod
    def create_genotyping(data: SeriesData) -> tuple[WellItem, WellItem]:
        identifier = data[int, "Well"]
        snp_name = data[
            str, "SNP Assay Name", f"Unable to find snp name for well {identifier}"
        ]
        allele1 = data[
            str, "Allele1 Name", f"Unable to find allele 1 for well {identifier}"
        ]
        allele2 = data[
            str, "Allele2 Name", f"Unable to find allele 2 for well {identifier}"
        ]
        return (
            WellItem(
                uuid=random_uuid_str(),
                identifier=identifier,
                target_dna_description=f"{snp_name}-{allele1}",
                sample_identifier=data.get(str, "Sample Name", NOT_APPLICABLE),
                reporter_dye_setting=data.get(str, "Allele1 Reporter"),
                position=data.get(str, "Well Position", NOT_APPLICABLE),
                location_identifier=data[str, "Well"],
                well_location_identifier=data.get(str, "Well Position"),
                quencher_dye_setting=data.get(str, "Quencher"),
                group_identifier=data.get(str, "Biogroup Name"),
                sample_role_type=SAMPLE_ROLE_TYPES_MAP.get(data.get(str, "Task", "")),
                extra_data={
                    "well identifier": identifier,
                    "sample color": data.get(str, "Sample Color"),
                    "biogroup color": data.get(str, "Biogroup Color"),
                    "target color": data.get(str, "Target Color"),
                },
            ),
            WellItem(
                uuid=random_uuid_str(),
                identifier=identifier,
                target_dna_description=f"{snp_name}-{allele2}",
                sample_identifier=data.get(str, "Sample Name", NOT_APPLICABLE),
                reporter_dye_setting=data.get(str, "Allele2 Reporter"),
                position=data.get(str, "Well Position", NOT_APPLICABLE),
                location_identifier=data[str, "Well"],
                well_location_identifier=data.get(str, "Well Position"),
                quencher_dye_setting=data.get(str, "Quencher"),
                group_identifier=data.get(str, "Biogroup Name"),
                sample_role_type=SAMPLE_ROLE_TYPES_MAP.get(data.get(str, "Task", "")),
                extra_data={
                    "well identifier": identifier,
                    "sample color": data.get(str, "Sample Color"),
                    "biogroup color": data.get(str, "Biogroup Color"),
                    "target color": data.get(str, "Target Color"),
                },
            ),
        )

    @staticmethod
    def create_generic(data: SeriesData) -> WellItem:
        identifier = data[int, "Well"]
        return WellItem(
            uuid=random_uuid_str(),
            identifier=identifier,
            target_dna_description=data[
                str,
                "Target Name",
                f"Unable to find target dna description for well {identifier}",
            ],
            sample_identifier=data.get(str, "Sample Name", NOT_APPLICABLE),
            reporter_dye_setting=data.get(str, "Reporter"),
            position=data.get(str, "Well Position", NOT_APPLICABLE),
            location_identifier=data[str, "Well"],
            well_location_identifier=data.get(str, "Well Position"),
            quencher_dye_setting=data.get(str, "Quencher"),
            group_identifier=data.get(str, "Biogroup Name"),
            sample_role_type=SAMPLE_ROLE_TYPES_MAP.get(data.get(str, "Task", "")),
            extra_data={
                "well identifier": identifier,
                "sample color": data.get(str, "Sample Color"),
                "biogroup color": data.get(str, "Biogroup Color"),
                "target color": data.get(str, "Target Color"),
            },
        )


@dataclass
class Well:
    identifier: int
    items: list[WellItem]

    @staticmethod
    def create_genotyping(data: SeriesData) -> Well:
        return Well(
            identifier=try_int(str(data.series.name), "well id"),
            items=list(WellItem.create_genotyping(data)),
        )

    @staticmethod
    def create_generic(well_data: pd.DataFrame) -> Well:
        return Well(
            identifier=try_int(str(well_data["Well"].iloc[0]), "well id"),
            items=[
                WellItem.create_generic(SeriesData(item_data))
                for _, item_data in well_data.iterrows()
            ],
        )

    @staticmethod
    def create(
        reader: AppBioQuantStudioReader, experiment_type: ExperimentType
    ) -> list[Well]:
        if (
            data := reader.sections.get("Sample Setup", reader.sections.get("Results"))
        ) is None:
            msg = "Expected 'Sample Setup' or 'Results' section"
            raise AllotropeConversionError(msg)

        if experiment_type == ExperimentType.genotyping_qpcr_experiment:
            return map_rows(data, Well.create_genotyping)
        else:
            return list(
                map_wells(
                    Well.create_generic, data[data["Target Name"].notnull()]
                ).values()
            )


@dataclass(frozen=True)
class AmplificationData:
    total_cycle_number_setting: float
    cycle: list[float]
    rn: list[float | None]
    delta_rn: list[float | None]


def create_amplification_data(
    reader: AppBioQuantStudioReader,
) -> dict[int, dict[str, AmplificationData]]:
    if (data := reader.sections.get("Amplification Data")) is None:
        return {}

    def make_data(well_data: pd.DataFrame) -> dict[str, AmplificationData]:
        return {
            str(target_name): AmplificationData(
                total_cycle_number_setting=float(target_data["Cycle"].max()),
                cycle=target_data["Cycle"].tolist(),
                rn=target_data["Rn"].tolist(),
                delta_rn=target_data["Delta Rn"].tolist(),
            )
            for target_name, target_data in well_data.groupby("Target Name")
        }

    return map_wells(make_data, data)


@dataclass(frozen=True)
class MulticomponentData:
    cycle: list[float]
    columns: dict[str, list[float | None]]

    def get_column(self, name: str) -> list[float | None]:
        return assert_not_none(
            self.columns.get(name),
            msg=f"Unable to obtain '{name}' from multicomponent data.",
        )


def create_multicomponent_data(
    reader: AppBioQuantStudioReader,
) -> dict[int, MulticomponentData]:
    if (data := reader.sections.get("Multicomponent Data")) is None:
        return {}

    def make_data(well_data: pd.Series[Any]) -> MulticomponentData:
        return MulticomponentData(
            cycle=well_data["Cycle"].tolist(),
            columns={
                name: well_data[name].tolist()
                for name in well_data
                if name not in ["Well", "Cycle", "Well Position"]
            },
        )

    return map_wells(make_data, data)


@dataclass(frozen=True)
class ResultMetadata:
    reference_dna_description: str | None
    reference_sample_description: str | None

    @staticmethod
    def create(data: SeriesData, experiment_type: ExperimentType) -> ResultMetadata:
        if experiment_type not in [
            ExperimentType.comparative_ct_qpcr_experiment,
            ExperimentType.relative_standard_curve_qpcr_experiment,
        ]:
            return ResultMetadata(None, None)
        return ResultMetadata(
            reference_dna_description=data.get(
                str, "Endogenous Control", NOT_APPLICABLE
            ),
            reference_sample_description=data.get(
                str, "Reference Sample", NOT_APPLICABLE
            ),
        )


@dataclass(frozen=True)
class Result:
    cycle_threshold_value_setting: float
    cycle_threshold_result: float | None
    automatic_cycle_threshold_enabled_setting: bool | None
    automatic_baseline: bool | None
    baseline_start: int | None
    baseline_end: int | None
    normalized_reporter_result: float | None
    baseline_corrected_reporter_result: float | None
    genotyping_determination_result: str | None
    genotyping_determination_method_setting: float | None
    quantity: float | None
    quantity_mean: float | None
    quantity_sd: float | None
    ct_mean: float | None
    ct_sd: float | None
    delta_ct_mean: float | None
    delta_ct_se: float | None
    delta_delta_ct: float | None
    rq: float | None
    rq_min: float | None
    rq_max: float | None
    rn_mean: float | None
    rn_sd: float | None
    y_intercept: float | None
    r_squared: float | None
    slope: float | None
    efficiency: float | None
    comments: str | None
    amp_score: float | None
    cq_conf: float | None
    extra_data: dict[str, Any] | None = None

    @staticmethod
    def create(
        reader: AppBioQuantStudioReader, experiment_type: ExperimentType
    ) -> tuple[dict[int, dict[str, Result]], ResultMetadata]:
        if (data := reader.sections.get("Results")) is None:
            msg = "Expected 'Results' section in file"
            raise AllotropeConversionError(msg)

        result = Result.create_results(data, experiment_type)
        metadata = SeriesData(pd.Series())
        if (raw_metadata := reader.sections.get("Results Metadata")) is not None:
            metadata = df_to_series_data(raw_metadata)

        return result, ResultMetadata.create(metadata, experiment_type)

    @staticmethod
    def create_results(
        data: pd.DataFrame, experiment_type: ExperimentType
    ) -> dict[int, dict[str, Result]]:
        target_key = "Target Name"
        if experiment_type == ExperimentType.genotyping_qpcr_experiment:
            target_key = "SNP Assay Name"

        def make_results(well_data: pd.DataFrame) -> dict[str, Result]:
            return {
                target_dna_description: result
                for target_id, target_data in well_data.groupby(target_key)
                for target_dna_description, result in Result.create_result(
                    df_to_series_data(target_data),
                    experiment_type,
                    str(target_id),
                ).items()
            }

        return map_wells(make_results, data)

    @staticmethod
    def create_result(
        data: SeriesData, experiment_type: ExperimentType, target_id: str
    ) -> dict[str, Result]:
        ct_col = Result.get_ct_col(list(data.series.index.astype(str)))
        ct_prefix = ct_col.capitalize()
        if experiment_type == ExperimentType.genotyping_qpcr_experiment:
            allele_prefixes = []
            for column in data.series.index:
                if match := re.match(rf"(^\w+) {ct_prefix}$", column):
                    allele_prefixes.append(f"{match.groups()[0]} ")
        else:
            allele_prefixes = [""]

        return {
            f"{target_id}{f'-{allele_prefix}' if allele_prefix else ''}".replace(
                " ", ""
            ): Result(
                cycle_threshold_value_setting=data[
                    float, f"{allele_prefix}{ct_prefix} Threshold"
                ],
                # TODO(nstender): really seems like this should be NaN if invalid value. Keeping to preserve tests.
                cycle_threshold_result=data.get(float, f"{allele_prefix}{ct_col}"),
                automatic_cycle_threshold_enabled_setting=data.get(
                    bool, f"{allele_prefix}Automatic {ct_prefix} Threshold"
                ),
                automatic_baseline=data.get(bool, f"{allele_prefix}Automatic Baseline"),
                baseline_start=data.get(int, f"{allele_prefix}Baseline Start"),
                baseline_end=data.get(int, f"{allele_prefix}Baseline End"),
                normalized_reporter_result=data.get(float, "Rn"),
                baseline_corrected_reporter_result=data.get(
                    float, f"{allele_prefix}Delta Rn"
                ),
                genotyping_determination_result=data.get(str, "Call"),
                genotyping_determination_method_setting=data.get(
                    float, "Threshold Value"
                ),
                quantity=data.get(float, "Quantity"),
                quantity_mean=data.get(float, "Quantity Mean"),
                quantity_sd=data.get(float, "Quantity SD"),
                ct_mean=data.get(float, f"{ct_prefix} Mean"),
                ct_sd=data.get(float, f"{ct_prefix} SD"),
                delta_ct_mean=data.get(float, f"Delta {ct_prefix} Mean"),
                delta_ct_se=data.get(float, f"Delta {ct_prefix} SE"),
                delta_delta_ct=data.get(float, f"Delta Delta {ct_prefix}"),
                rq=data.get(float, "RQ"),
                rq_min=data.get(float, "RQ Min"),
                rq_max=data.get(float, "RQ Max"),
                rn_mean=data.get(float, "Rn Mean"),
                rn_sd=data.get(float, "Rn SD"),
                y_intercept=data.get(float, "Y-Intercept"),
                r_squared=data.get(float, "R(superscript 2)"),
                slope=data.get(float, "Slope"),
                efficiency=data.get(float, "Efficiency"),
                comments=data.get(str, "Comments"),
                amp_score=data.get(float, "Amp Score"),
                cq_conf=data.get(float, "Cq Conf"),
                extra_data={
                    "omit": data.get(bool, "Omit"),
                    "highsd": data.get(str, "HIGHSD"),
                    "noamp": data.get(str, "NOAMP"),
                    "expfail": data.get(str, "EXPFAIL"),
                    "tholdfail": data.get(str, "THOLDFAIL"),
                    "prfdrop": data.get(str, "PRFDROP"),
                },
            )
            for allele_prefix in allele_prefixes
        }

    @staticmethod
    def get_ct_col(columns: list[str]) -> str:
        """Looks for any column matching the pattern of the cycle threshhold.

        The pattern for the threshold column is `[allele prefix]<quantification_method>`
        where `allele prefix` is optional and `quantification_method` can be any of
        Ct, Crt or Cq with different capitalizations.

        This identifier is used as a prefix to all the cycle threshold related
        columns (always capitalized)
        """
        r = re.compile(r".*(C[tT]|C(?:rt|RT)|C[qQ])$")
        for column in columns:
            if match := r.match(column):
                return match.groups()[0]

        msg = "Unable to identify ct prefix"
        raise AllotropeConversionError(msg)


@dataclass(frozen=True)
class MeltCurveRawData:
    reading: list[float]
    fluorescence: list[float | None]
    derivative: list[float | None]

    @staticmethod
    def create(reader: AppBioQuantStudioReader) -> dict[int, MeltCurveRawData]:
        if (data := reader.sections.get("Melt Curve Raw Data")) is None:
            return {}

        def make_data(well_data: pd.Series[Any]) -> MeltCurveRawData:
            return MeltCurveRawData(
                reading=well_data["Reading"].tolist(),
                fluorescence=well_data["Fluorescence"].tolist(),
                derivative=well_data["Derivative"].tolist(),
            )

        return map_wells(make_data, data)



import json
from pathlib import Path
from typing import Any


def unit_name_from_iri(unit_iri: str) -> str:
    return Path(unit_iri).name.split("#")[-1]


UNIT_STRING_OVERRIDES = {"(unitless)": "UNITLESS"}


def _update_unit_models(
    units_schema: dict[str, Any],
    units_models_path: Path,
    custom_models_path: Path,
) -> None:
    with open(units_models_path, "w") as f:
        f.write(
            """# THIS IS AN AUTOGENERATED FILE. DO NOT EDIT THIS FILE DIRECTLY.
from dataclasses import dataclass

"""
        )
        for unit_name, override in UNIT_STRING_OVERRIDES.items():
            f.write(f'{override} = "{unit_name}"')

        f.write(
            """
@dataclass(frozen=True, kw_only=True)
class HasUnit:
    unit: str
"""
        )
        for unit_name, unit_schema in units_schema.items():
            unit = unit_schema["properties"]["unit"]["const"]
            unit_str = f"'{unit}'" if '"' in unit else f'"{unit}"'
            f.write(
                f"""


@dataclass(frozen=True, kw_only=True)
class {unit_name}(HasUnit):
    unit: str = {UNIT_STRING_OVERRIDES.get(unit, unit_str)}"""
            )

    with open(custom_models_path, "w") as f:
        f.write(
            """# THIS IS AN AUTOGENERATED FILE. DO NOT EDIT THIS FILE DIRECTLY.
from dataclasses import dataclass

from allotropy.allotrope.models.shared.definitions.definitions import (
    TNullableQuantityValue,
    TQuantityValue,
)
from allotropy.allotrope.models.shared.definitions.units import (
"""
        )
        for unit_name in units_schema:
            f.write(f"    {unit_name},\n")
        f.write(")\n")

        for unit_name in units_schema:
            f.write(
                f"""

@dataclass(frozen=True, kw_only=True)
class TQuantityValue{unit_name}({unit_name}, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValue{unit_name}({unit_name}, TNullableQuantityValue):
    pass
"""
            )


def _get_unit_schema(unit: str, unit_iri: str) -> dict[str, Any]:
    return {
        "properties": {
            "unit": {"type": "string", "const": unit, "$asm.unit-iri": unit_iri}
        },
        "required": ["unit"],
    }


def _get_quantity_value_schema(
    unit_name: str, prefix: str | None = ""
) -> dict[str, Any]:
    return {
        "allOf": [
            {"$ref": f"#/$defs/t{prefix}QuantityValue"},
            {"$ref": f"#/$defs/{unit_name}"},
        ]
    }


def update_unit_files(
    unit_to_iri: dict[str, str],
    unit_schemas_path: Path,
    units_models_path: Path,
    custom_schemas_path: Path,
    custom_models_path: Path,
) -> None:
    with open(unit_schemas_path) as f:
        units_schema = json.load(f)

    for unit, unit_iri in unit_to_iri.items():
        units_schema[unit_name_from_iri(unit_iri)] = _get_unit_schema(unit, unit_iri)

    units_schema = dict(sorted(units_schema.items()))
    with open(unit_schemas_path, "w") as f:
        json.dump(units_schema, f, indent=2, ensure_ascii=False)

    with open(unit_schemas_path) as f:
        units_schema = json.load(f)

    custom_schema = {}
    for unit_name in units_schema:
        custom_schema[f"tQuantityValue{unit_name}"] = _get_quantity_value_schema(
            unit_name
        )
        custom_schema[
            f"tNullableQuantityValue{unit_name}"
        ] = _get_quantity_value_schema(unit_name, "Nullable")

    with open(custom_schemas_path, "w") as f:
        json.dump(custom_schema, f, indent=2, ensure_ascii=False)

    _update_unit_models(units_schema, units_models_path, custom_models_path)



from collections import defaultdict

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.mabtech_apex.mabtech_apex_reader import MabtechApexReader
from allotropy.parsers.mabtech_apex.mabtech_apex_structure import (
    create_measurement_group,
    create_metadata,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.pandas import (
    SeriesData,
)
from allotropy.parsers.vendor_parser import VendorParser


class MabtechApexParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Mabtech Apex"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = MabtechApexReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = MabtechApexReader.create(named_file_contents)

        # if Read Date or machine ID is not present in file, return None, no measurement for given Well
        plate_data = reader.data.dropna(subset=["Read Date", "Machine ID"])
        measurements_per_well = defaultdict(list)
        for _, row in plate_data.iterrows():
            well_row = SeriesData(row)
            well = well_row.get(str, "Well")
            measurements_per_well[well].append(well_row)

        data_measurement = [
            create_measurement_group(measurements, reader.plate_info)
            for measurements in measurements_per_well.values()
        ]
        data = Data(
            create_metadata(reader.plate_info, named_file_contents.original_file_path),
            data_measurement,
        )
        return data



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    CalculatedDataItem,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
    ProcessedDataFeature,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.thermo_fisher_nanodrop_8000.nanodrop_8000_structure import (
    create_calculated_data,
    read_absorbances,
    read_mass_concentration_capture_wavelength,
)
from allotropy.parsers.thermo_fisher_nanodrop_eight import constants
from allotropy.parsers.utils.iterables import get_first_not_none
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import try_float, try_float_or_none


@dataclass
class SpectroscopyRow:
    timestamp: str
    experiment_type: str | None
    measurements: list[Measurement]
    calculated_data: list[CalculatedDataItem]

    @staticmethod
    def create(data: SeriesData, header: SeriesData) -> SpectroscopyRow:
        absorbances = read_absorbances(data)
        experiment_type = header.get(str, "application")
        mass_concentration_capture_wavelength = (
            read_mass_concentration_capture_wavelength(
                data, experiment_type, absorbances
            )
        )
        mass_concentration = get_first_not_none(
            lambda key: data.get(float, key),
            ["conc.", "conc", "concentration"],
        )
        unit = get_first_not_none(
            lambda key: key if data.get(float, key.lower()) else None,
            constants.CONCENTRATION_UNITS,
        )

        spectra_data_cube = None
        float_cols = [col for col in data.series.index if try_float_or_none(col)]
        spectra_data = {try_float(col, col): data.get(float, col) for col in float_cols}
        if len(spectra_data) and None not in spectra_data.values():
            spectra_data_cube = DataCube(
                label="absorption spectrum",
                structure_dimensions=[
                    DataCubeComponent(FieldComponentDatatype.double, "wavelength", "nm")
                ],
                structure_measures=[
                    DataCubeComponent(
                        FieldComponentDatatype.double, "absorbance", "mAU"
                    ),
                ],
                dimensions=[list(spectra_data.keys())],
                measures=[list(spectra_data.values())],
            )

        sample_id = data.get(
            str, ["sample id", "uid"], NOT_APPLICABLE, SeriesData.NOT_NAN
        )
        location_id = data.get(str, "location")
        measurements: list[Measurement] = []
        for wavelength, absorbance in absorbances.items():
            if not absorbance:
                continue
            measurements.append(
                Measurement(
                    type_=MeasurementType.ULTRAVIOLET_ABSORBANCE,
                    identifier=random_uuid_str(),
                    data_cube=spectra_data_cube,
                    absorbance=absorbance,
                    detector_wavelength_setting=wavelength,
                    sample_identifier=sample_id,
                    location_identifier=location_id,
                    processed_data=ProcessedData(
                        features=[
                            ProcessedDataFeature(
                                result=mass_concentration,
                                unit=unit,
                            )
                        ],
                    )
                    if mass_concentration_capture_wavelength == wavelength
                    and mass_concentration
                    and unit
                    else None,
                )
            )
        if spectra_data_cube:
            measurements.append(
                Measurement(
                    type_=MeasurementType.ULTRAVIOLET_ABSORBANCE_SPECTRUM,
                    identifier=random_uuid_str(),
                    data_cube=spectra_data_cube,
                    sample_identifier=sample_id,
                    location_identifier=location_id,
                )
            )

        calculated_data = create_calculated_data(data, measurements)

        return SpectroscopyRow(
            data[str, ["date", "date & time"]],
            experiment_type,
            measurements,
            calculated_data,
        )


def create_metadata(data: SeriesData, file_path: str) -> Metadata:
    return Metadata(
        device_identifier=constants.DEVICE_IDENTIFIER,
        device_type=constants.DEVICE_TYPE,
        model_number=constants.MODEL_NUBMER,
        equipment_serial_number=data[str, "serial number"],
        file_name=Path(file_path).name,
        unc_path=file_path,
    )


def create_measurement_group(
    row: SpectroscopyRow, header: SeriesData
) -> MeasurementGroup:
    return MeasurementGroup(
        measurement_time=row.timestamp,
        analyst=header.get(str, "user name"),
        experiment_type=row.experiment_type,
        measurements=row.measurements,
    )



from collections import defaultdict
from copy import deepcopy
from pathlib import Path

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.pcr.rec._2024._09.qpcr import (
    Error,
    Measurement,
    MeasurementGroup,
    Metadata,
    ProcessedData,
)
from allotropy.parsers.cfxmaestro import constants
from allotropy.parsers.constants import (
    DEFAULT_EPOCH_TIMESTAMP,
    NEGATIVE_ZERO,
    NOT_APPLICABLE,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_metadata(file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        file_name=path.name,
        asm_file_identifier=path.with_suffix(".json").name,
        unc_path=file_path,
        experiment_type=constants.EXPERIMENT_TYPE,
        container_type=constants.CONTAINER_TYPE,
        software_name=constants.SOFTWARE_NAME,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        device_type=constants.DEVICE_TYPE,
        device_serial_number=NOT_APPLICABLE,
        device_identifier=NOT_APPLICABLE,
        data_system_instance_identifier=NOT_APPLICABLE,
        model_number=NOT_APPLICABLE,
        measurement_method_identifier=NOT_APPLICABLE,
    )


def create_measurement_group(
    well_data: list[SeriesData],
    plate_well_count: int,
) -> MeasurementGroup:
    return MeasurementGroup(
        plate_well_count=plate_well_count,
        well_volume=NEGATIVE_ZERO,
        error_document=[
            Error(error="Value not provided in instrument file", feature="well volume")
        ],
        experimental_data_identifier=NOT_APPLICABLE,
        measurements=[
            Measurement(
                # Measurement metadata
                identifier=random_uuid_str(),
                sample_identifier=data.get(
                    str, "Sample", NOT_APPLICABLE, SeriesData.NOT_NAN
                ),
                target_identifier=data.get(
                    str, "Target", NOT_APPLICABLE, SeriesData.NOT_NAN
                ),
                group_identifier=data.get(
                    str, "Biological Set Name", NOT_APPLICABLE, SeriesData.NOT_NAN
                ),
                timestamp=DEFAULT_EPOCH_TIMESTAMP,
                # Optional measurement metadata
                sample_role_type=constants.SAMPLE_ROLE_TYPES_MAP.get(
                    data.get(str, "Content", "__INVALID_KEY__")
                ),
                location_identifier=data[str, "Well"],
                well_location_identifier=data[str, "Well"],
                # Optional settings
                reporter_dye_setting=data[str, "Fluor"],
                # Processed data
                processed_data=ProcessedData(
                    # TODO: add add error document (or omit?) if Cq is NaN.
                    cycle_threshold_result=data.get(
                        float, "Cq", validate=SeriesData.NOT_NAN
                    ),
                    # TODO: cycle number is required, but we do not have any examples of the value being
                    # provided, if we get one, and the column does not match, update here.
                    cycle_threshold_value_setting=(
                        cycle_number := data.get(float, "Cycle Number", NEGATIVE_ZERO)
                    ),
                ),
                # Since the processed data doc does not include an error document,
                # we added the cycle threshold value setting error at measurement level
                error_document=(
                    [
                        Error(
                            error="Value not provided in instrument file",
                            feature="cycle threshold value setting (qPCR)",
                        )
                    ]
                    if cycle_number == NEGATIVE_ZERO
                    else None
                ),
            )
            for data in well_data
            if data.get(str, "Sample", validate=SeriesData.NOT_NAN)
            or data.get(float, "Cq", validate=SeriesData.NOT_NAN)
        ],
    )


def create_measurement_groups(df: pd.DataFrame) -> list[MeasurementGroup]:
    well_to_rows = defaultdict(list)

    def map_to_dict(data: SeriesData) -> None:
        well_to_rows[data[str, "Well"]].append(deepcopy(data))

    map_rows(df, map_to_dict)

    groups = [
        create_measurement_group(well_to_rows[well_id], len(well_to_rows))
        for well_id in well_to_rows
    ]
    return [group for group in groups if group.measurements]



from pathlib import Path

import pandas as pd

from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Data,
    Mapper,
)
from allotropy.constants import DEFAULT_ENCODING
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.thermo_fisher_nanodrop_one.constants import DISPLAY_NAME
from allotropy.parsers.thermo_fisher_nanodrop_one.thermo_fisher_nanodrop_one_structure import (
    DataNanodrop,
)
from allotropy.parsers.utils.pandas import read_csv, read_multisheet_excel
from allotropy.parsers.vendor_parser import VendorParser


class ThermoFisherNanodropOneParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "csv,xlsx"
    SCHEMA_MAPPER = Mapper

    def read_data(
        self, named_file_contents: NamedFileContents
    ) -> tuple[pd.DataFrame, str]:
        if named_file_contents.extension == "xlsx":
            xlsx_data = read_multisheet_excel(
                named_file_contents.contents,
                engine="calamine",
            )
            sheet_name = next(iter(xlsx_data.keys()))
            experiment_type, *_ = sheet_name.split(maxsplit=1)
            return xlsx_data[sheet_name], experiment_type

        csv_data = read_csv(
            named_file_contents.contents,
            encoding=DEFAULT_ENCODING,
        )
        file_name = Path(named_file_contents.original_file_path).name
        experiment_type, *_ = file_name.split(maxsplit=1)
        return csv_data, experiment_type

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        data, experiment_type = self.read_data(named_file_contents)
        return DataNanodrop.create(
            data, experiment_type, named_file_contents.original_file_path
        )



from allotropy.allotrope.models.adm.pcr.rec._2024._09.qpcr import Model
from allotropy.allotrope.schema_mappers.adm.pcr.rec._2024._09.qpcr import Data, Mapper
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.cfxmaestro.cfxmaestro_reader import (
    CFXMaestroReader,
)
from allotropy.parsers.cfxmaestro.cfxmaestro_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.cfxmaestro.constants import DISPLAY_NAME
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class CfxmaestroParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.WORKING_DRAFT
    SUPPORTED_EXTENSIONS = CFXMaestroReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = CFXMaestroReader(named_file_contents)
        return Data(
            create_metadata(named_file_contents.original_file_path),
            create_measurement_groups(reader.data),
        )



from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType

DISPLAY_NAME = "BMG Labtech SMART Control"
PRODUCT_MANUFACTURER = "BMG LABTECH"
SOFTWARE_NAME = "SMART Control"
DEVICE_TYPE = "fluorescence detector"

TARGET_TEMPERATURE = "Target temperature [°C]:"
SAMPLE_ROLE_TYPE_MAPPING = {
    "Blank": SampleRoleType.blank_role,
    "Standard": SampleRoleType.standard_sample_role,
    "Sample": SampleRoleType.sample_role,
}



from abc import abstractmethod

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMilliliter,
    TQuantityValuePercent,
    TQuantityValueSeimensPerMeter,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
    TQuantityValue,
)
from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    DeviceControlDoc,
    Peak,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCubeComponent
from allotropy.parsers.cytiva_unicorn.constants import (
    DEVICE_TYPE,
    PEAK_AREA_UNIT,
    PEAK_END_UNIT,
    PEAK_START_UNIT,
    PEAK_WIDTH_AT_HALF_HEIGHT_UNIT,
    PEAK_WIDTH_UNIT,
)
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.cytiva_unicorn.structure.measurements.generic import (
    UnicornMeasurement,
)
from allotropy.parsers.cytiva_unicorn.structure.static_docs import (
    StaticDocs,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none, quantity_or_none


class AbsorbanceMeasurement(UnicornMeasurement):
    @classmethod
    @abstractmethod
    def get_curve_regex(cls) -> str:
        pass

    @classmethod
    def get_peaks(cls, _: UnicornZipHandler) -> list[Peak]:
        return []

    @classmethod
    def create_or_none(
        cls,
        handler: UnicornZipHandler,
        elements: list[StrictXmlElement],
        static_docs: StaticDocs,
    ) -> UnicornMeasurement:
        return cls.get_measurement(
            static_docs=static_docs,
            chromatogram_data_cube=assert_not_none(
                cls.get_data_cube_or_none(
                    handler,
                    cls.filter_curve_or_none(elements, cls.get_curve_regex()),
                    DataCubeComponent(
                        type_=FieldComponentDatatype.float,
                        concept="absorbance",
                        unit="mAU",
                    ),
                ),
                msg="Unable to find information to create absorbance data cubes.",
            ),
            device_control_docs=[
                DeviceControlDoc(
                    device_type=DEVICE_TYPE,
                    start_time=static_docs.start_time,
                )
            ],
            peaks=cls.get_peaks(handler),
        )


class AbsorbanceMeasurement1(AbsorbanceMeasurement):
    @classmethod
    def get_curve_regex(cls) -> str:
        return r"^UV 1_\d+$"

    @classmethod
    def get_peaks_custom_info(
        cls, peak: StrictXmlElement
    ) -> dict[str, TQuantityValue | None]:
        start_conduct_height = peak.get_sub_float_or_none("StartConductivityHeight")
        max_conductivity_height = peak.get_sub_float_or_none("MaxConductivityHeight")
        end_conductivity_height = peak.get_sub_float_or_none("EndConductivityHeight")
        average_conductivity = peak.get_sub_float_or_none("AverageConductivity")

        return {
            "max peak retention": quantity_or_none(
                TQuantityValueMilliliter,
                peak.get_sub_float_or_none("MaxPeakRetention"),
            ),
            "percent of total area": quantity_or_none(
                TQuantityValuePercent,
                peak.get_sub_float_or_none("PercentOfTotalArea"),
            ),
            "start peak end point height": quantity_or_none(
                TQuantityValueMilliAbsorbanceUnit,
                peak.get_sub_float_or_none("StartPeakEndpointHeight"),
            ),
            "end peak end point height": quantity_or_none(
                TQuantityValueMilliAbsorbanceUnit,
                peak.get_sub_float_or_none("EndPeakEndpointHeight"),
            ),
            "sigma": quantity_or_none(
                TQuantityValueUnitless,
                peak.get_sub_float_or_none("Sigma"),
            ),
            "asymmetry peak start": quantity_or_none(
                TQuantityValueUnitless,
                peak.get_sub_float_or_none("AssymetryPeakStart"),
            ),
            "asymmetry peak end": quantity_or_none(
                TQuantityValueUnitless,
                peak.get_sub_float_or_none("AssymetryPeakEnd"),
            ),
            "start conductivity height": (
                TQuantityValueSeimensPerMeter(value=start_conduct_height / 10)
                if start_conduct_height
                else None
            ),
            "max conductivity height": (
                TQuantityValueSeimensPerMeter(value=max_conductivity_height / 10)
                if max_conductivity_height
                else None
            ),
            "end conductivity height": (
                TQuantityValueSeimensPerMeter(value=end_conductivity_height / 10)
                if end_conductivity_height
                else None
            ),
            "average conductivity": (
                TQuantityValueSeimensPerMeter(value=average_conductivity / 10)
                if average_conductivity
                else None
            ),
        }

    @classmethod
    def get_peaks(cls, handler: UnicornZipHandler) -> list[Peak]:
        chrom_1 = handler.get_chrom_1()
        peaks = chrom_1.recursive_find_or_none(["PeakTables", "PeakTable", "Peaks"])
        return [
            Peak(
                identifier=random_uuid_str(),
                index=f"Peak {idx}",
                end=peak.get_sub_float_or_none("EndPeakRetention"),
                end_unit=PEAK_END_UNIT,
                height=peak.get_sub_float_or_none("Height"),
                written_name=peak.get_sub_text_or_none("Name"),
                area=peak.get_sub_float_or_none("Area"),
                area_unit=PEAK_AREA_UNIT,
                relative_area=peak.get_sub_float_or_none("PercentOfTotalPeakArea"),
                start=peak.get_sub_float_or_none("StartPeakRetention"),
                start_unit=PEAK_START_UNIT,
                chromatographic_resolution=peak.get_sub_float_or_none("Resolution"),
                width_at_half_height=peak.get_sub_float_or_none("WidthAtHalfHeight"),
                width_at_half_height_unit=PEAK_WIDTH_AT_HALF_HEIGHT_UNIT,
                width=peak.get_sub_float_or_none("Width"),
                width_unit=PEAK_WIDTH_UNIT,
                chromatographic_asymmetry=peak.get_sub_float_or_none("Assymetry"),
                custom_info=cls.get_peaks_custom_info(peak),
            )
            for idx, peak in enumerate(peaks.findall("Peak") if peaks else [], start=1)
        ]


class AbsorbanceMeasurement2(AbsorbanceMeasurement):
    @classmethod
    def get_curve_regex(cls) -> str:
        return r"^UV 2_\d+$"


class AbsorbanceMeasurement3(AbsorbanceMeasurement):
    @classmethod
    def get_curve_regex(cls) -> str:
        return r"^UV 3_\d+$"



from allotropy.allotrope.models.adm.plate_reader.rec._2024._06.plate_reader import Model
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.moldev_softmax_pro.softmax_pro_data_creator import (
    create_calculated_data,
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.moldev_softmax_pro.softmax_pro_structure import (
    StructureData,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class SoftmaxproParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Molecular Devices SoftMax Pro"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "txt"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        lines = read_to_lines(named_file_contents)
        reader = CsvReader(lines)
        data = StructureData.create(reader)

        return Data(
            create_metadata(named_file_contents.original_file_path),
            create_measurement_groups(data),
            create_calculated_data(data),
        )



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.methodical_mind import constants
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str

WELL_LABELS = ["A", "B", "C", "D", "E", "F", "G", "H"]


@dataclass(frozen=True)
class Header:
    file_name: str
    version: str
    model: str
    serial_number: str

    @staticmethod
    def create(data: SeriesData) -> Header:
        return Header(
            file_name=data[str, "FileName"],
            version=data[str, "Version"],
            model=data[str, "Model"],
            serial_number=data[str, "Serial #"],
        )


@dataclass(frozen=True)
class PlateData:
    measurement_time: str
    plate_well_count: int
    analyst: str | None
    well_plate_id: str
    well_data: list[WellData]

    @staticmethod
    def create(
        header: SeriesData,
        data: pd.DataFrame,
    ) -> PlateData:
        well_plate_id = header[str, "Barcode1"].strip("<>")
        unique_well_labels = [
            label for label in data.index.unique() if label in WELL_LABELS
        ]
        well_data = [
            WellData.create(
                luminescence=value,
                location_id=str(row_index + 1),
                well_plate_id=well_plate_id,
                well_location_id=f"{row_name}{col_name}",
            )
            # Get each unique row label, and then iterate over all rows with that label.
            for row_name in unique_well_labels
            for row_index, (_, row) in enumerate(data.loc[[row_name]].iterrows())
            for col_name, value in row.items()
            if row_name in WELL_LABELS
        ]
        return PlateData(
            measurement_time=header[str, "Read Time"],
            analyst=header.get(str, "User"),
            well_plate_id=well_plate_id,
            # The well count is (# of unique row labels) * (# of columns)
            plate_well_count=len(unique_well_labels) * data.shape[1],
            well_data=well_data,
        )


@dataclass(frozen=True)
class WellData:
    luminescence: int
    location_identifier: str
    sample_identifier: str
    well_location_identifier: str

    @staticmethod
    def create(
        luminescence: int, location_id: str, well_plate_id: str, well_location_id: str
    ) -> WellData:
        sample_id = well_plate_id + "_" + well_location_id
        return WellData(
            luminescence=luminescence,
            location_identifier=location_id,
            sample_identifier=sample_id,
            well_location_identifier=well_location_id,
        )


def create_metadata(header: Header, file_name: str) -> Metadata:
    asm_file_identifier = Path(file_name).with_suffix(".json")
    return Metadata(
        file_name=header.file_name.rsplit("\\", 1)[-1],
        unc_path=header.file_name,
        software_name=header.version,
        software_version=header.version,
        device_identifier=NOT_APPLICABLE,
        model_number=header.model,
        equipment_serial_number=header.serial_number,
        asm_file_identifier=asm_file_identifier.name,
        data_system_instance_id=NOT_APPLICABLE,
    )


def create_measurement_groups(plates: list[PlateData]) -> list[MeasurementGroup]:
    plates_data = []
    for plate in plates:
        grouped_wells = defaultdict(list)
        for well in plate.well_data:
            grouped_wells[well.well_location_identifier].append(well)
        plates_data.extend(
            [
                MeasurementGroup(
                    analyst=plate.analyst,
                    measurement_time=plate.measurement_time,
                    plate_well_count=plate.plate_well_count,
                    measurements=[
                        Measurement(
                            type_=MeasurementType.LUMINESCENCE,
                            identifier=random_uuid_str(),
                            luminescence=well.luminescence,
                            sample_identifier=well.sample_identifier,
                            location_identifier=well.location_identifier,
                            well_location_identifier=well.well_location_identifier,
                            well_plate_identifier=plate.well_plate_id,
                            device_type=constants.LUMINESCENCE_DETECTOR,
                            detection_type=constants.LUMINESCENCE,
                        )
                        for well in well_group
                    ],
                )
                for well_group in grouped_wells.values()
            ]
        )

    return plates_data



from __future__ import annotations

from collections import defaultdict

from allotropy.calcdocs.extractor import Element
from allotropy.calcdocs.view import View


class ViewWithReference(View):
    def __init__(
        self,
        name: str,
        sub_view: ViewWithReference | None,
        reference: str | None = None,
    ):
        super().__init__(name, sub_view)
        self.reference = reference

    def filter_keys(self, keys: dict[str, str]) -> dict[str, str]:
        filtered_keys = super().filter_keys(keys)
        if self.reference is not None and self.name in filtered_keys:
            filtered_keys[self.name] = self.reference
        return filtered_keys


class SampleView(ViewWithReference):
    def __init__(
        self,
        sub_view: ViewWithReference | None = None,
        reference: str | None = None,
    ):
        super().__init__(name="sample_id", sub_view=sub_view, reference=reference)

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            if sample_identifier := element.get_str("sample_identifier"):
                items[str(sample_identifier)].append(element)
        return dict(items)


class TargetView(ViewWithReference):
    def __init__(
        self,
        sub_view: ViewWithReference | None = None,
        reference: str | None = None,
        blacklist: list[str] | None = None,
    ):
        super().__init__(name="target_dna", sub_view=sub_view, reference=reference)
        self.blacklist = blacklist

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            if target_dna := element.get_str("target_dna_description"):
                if self.blacklist is None or target_dna not in self.blacklist:
                    items[str(target_dna)].append(element)
        return dict(items)


class UuidView(ViewWithReference):
    def __init__(self, sub_view: ViewWithReference | None = None):
        super().__init__(name="uuid", sub_view=sub_view)

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            if uuid := element.get_str("uuid"):
                items[str(uuid)].append(element)
        return dict(items)


class TargetRoleView(ViewWithReference):
    def __init__(self, sub_view: ViewWithReference | None = None):
        super().__init__(name="target_dna", sub_view=sub_view)

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        items = defaultdict(list)
        for element in elements:
            if target_dna := element.get_str_or_none("target_dna_description"):
                if element.get_str("sample_role_type") == "standard sample role":
                    items[str(target_dna)].append(element)
        return dict(items)
